import asyncio
import time
from hedra.runners.leader_services.proto.leader_pb2 import PollLeaderJobRequest
from operator import countOf
from hedra.core import Executor
from hedra.reporting import Handler
from hedra.runners.leader_services.proto import PipelineStageRequest
from hedra.runners.utils.connect_timeout import connect_or_return_none

from hedra.core.personas.parallel import parse_time
from .worker_pipeline_config import WorkerPipelineConfig


class WorkerPipeline:

    def __init__(self, config, worker_address=None, worker_id=None, timeout=300) -> None:
        logger = Logger()
        self.session_logger = logger.generate_logger()

        self.distributed_config = config.distributed_config
        self.config = WorkerPipelineConfig()
        self.config.distributed_config = config.distributed_config
        self.reporter_config = None
        self.executor = None
        self.handler = None
        self.worker_address = worker_address
        self.worker_id = worker_id
        self.stage_completed = False
        self.leaders = []
        self.aggregate_events = []
        self.status = None
        self.timeout = timeout

    async def setup_pipeline(self, job, reporter_config, leaders):
        await self._parse_job_executor_config(job.get('executor_config'))
        actions = job.get('actions')

        self.timeout = job.get(
            'timeout', 
            parse_time(
                self.config.executor_config.get('total_time', '00:01:00')
            ) * 10
        )

        self.leaders = leaders
        self.reporter_config = reporter_config
        self.config.actions = actions

        self.executor = Executor(self.config)
        self.handler = Handler(self.config)

        self._worker_poll_rate = job.get('poll_rate', 1)

        await self.handler.initialize_reporter()
        self.status = 'created'

    async def run(self, job_id=None):
        try:

            self.status = 'active'

            await self.executor.setup(self.reporter_config)

            for pipeline_stage in self.executor.pipeline:
                await self.executor.pipeline.execute_stage(pipeline_stage)
                await self.mark_stage_completed(pipeline_stage.name, job_id=job_id)
                await self.wait_for_stage_completed(job_id=job_id)
            
            await self.executor.pipeline.get_results()

            completed_actions = self.executor.pipeline.stats.get('completed_actions')
            self.executor.session_logger.info(f'Processing - {completed_actions} - action results.')

            self.aggregate_events = await self.handler.aggregate(
                self.executor.pipeline.results
            )

            await self.mark_stage_completed('done', job_id=job_id)

        except Exception as err:

            self.status = 'failed'

            self.session_logger.error(f'\nEncountered exception running pipeline - {str(err)}\n')
            await self.mark_stage_completed('done', job_id=job_id)

            return self.handler.reporter.fields

        if job_id:
            self.session_logger.info(f'\nJob - {job_id} - has completed on worker - {str(self.worker_id)}\n')

        else:
            self.session_logger.info(f'Pipeline has completed.')

        self.status = 'completed'
        
        return {
            'aggregated': self.aggregate_events
        }

    async def check_for_completion(self, job_id=None):
        
        completions = []
        
        if len(self.leaders) > 0:
            completions, _ = await asyncio.wait([
                self._check_for_completion(
                    leader, 
                    job_id=job_id
                ) for leader in self.leaders
            ], timeout=self.timeout)

            completions = await asyncio.gather(*completions)

        self.stage_completed = countOf(completions, True) == len(self.leaders)

    @connect_or_return_none(wait_buffer=5, timeout_threshold=15)
    async def _check_for_completion(self, leader, job_id=None):
        request = PollLeaderJobRequest(
            host_address=self.worker_address,
            host_id=str(self.worker_id),
            job_id=job_id
        )

        response = await leader.client.PollLeaderJob(request)

        return response.completed

    async def mark_stage_completed(self, stage_name, job_id=None):
        if len(self.leaders) > 0:
            await asyncio.wait([
                self._mark_stage_completed(
                    leader,
                    stage_name,
                    job_id=job_id
                ) for leader in self.leaders
            ], timeout=self.timeout)


    @connect_or_return_none(wait_buffer=5, timeout_threshold=15)
    async def _mark_stage_completed(self, leader, stage_name, job_id=None):
        request = PipelineStageRequest(
            host_address=self.worker_address,
            host_id=str(self.worker_id),
            job_id=job_id,
            stage_completed=stage_name
        )

        return await leader.client.MarkStageCompleted(request)


    async def wait_for_stage_completed(self, job_id=None):
        elapsed = 0
        start = time.time()
        while self.stage_completed is False and elapsed < self.timeout:
            await self.check_for_completion(job_id=job_id)
            await asyncio.sleep(self._worker_poll_rate)
            elapsed = time.time() - start

        self.stage_completed = False

    async def _parse_job_executor_config(self, executor_config):
        executor_config['batch_size'] = int(executor_config.get('batch_size', 1000))
        executor_config['batch_count'] = int(executor_config.get('batch_count', 0))
        executor_config['warmup'] = int(executor_config.get('warmup', 0))
        executor_config['optimize'] = int(executor_config.get('optimize', 0))
        executor_config['batch_time'] = int(executor_config.get('batch_time', 10))
        executor_config['total_time'] = executor_config.get('total_time', '00:01:00')
        executor_config['optimize_iter_duration'] = int(executor_config.get('optimize_iter_duration', 0))
        executor_config['pool_size'] = int(executor_config.get('pool_size', 1))
        executor_config['no_run_visuals'] = True

        self.config.executor_config = executor_config

    async def update_leaders(self, active_leaders, failed_leaders, leaders):
        leaders = [
            leaders[leader] for leader in active_leaders if leader not in failed_leaders
        ]

        self.leaders = leaders



from .action_hook import ActionHook
from .after_hook import AfterHook
from .before_hook import BeforeHook
from .channel_hook import ChannelHook
from .check_hook import CheckHook
from .context_hook import ContextHook
from .event_hook import EventHook
from .internal_hook import InternallHook
from .metric_hook import MetricHook
from .restore_hook import RestoreHook
from .save_hook import SaveHook
from .setup_hook import SetupHook
from .task_hook import TaskHook
from .teardown_hook import TeardownHook
from .validate_hook import ValidateHook


import time
import asyncio
import psutil
import uuid
from typing import Dict, List, Union
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.tools.helpers import awaitable
from asyncio import Task
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.personas.batching.batch import Batch
from hedra.core.graphs.hooks.registry.registry_types import (
    ActionHook,
    TaskHook
)
from hedra.core.personas.batching import Batch
from hedra.core.engines.client.config import Config
from hedra.core.personas.types.types import PersonaTypes


async def cancel_pending(pend: Task):
    try:
        pend.cancel()
        if not pend.cancelled():
            await pend

        return pend
    
    except asyncio.CancelledError as cancelled_error:
        return cancelled_error


class DefaultPersona:

    __slots__ = (
        'metadata_string',
        'persona_id',
        'logger',
        'type',
        'workers',
        'actions',
        '_hooks',
        'batch',
        '_live_updates',
        'total_time',
        'duration',
        'total_actions',
        'total_elapsed',
        'start',
        'end',
        'completed_actions',
        'pending_actions',
        'completed_time',
        'run_timer',
        'actions_count',
        'graceful_stop',
        'is_timed',
        'timer_thread',
        'loop',
        'current_action_idx',
        'optimized_params',
        'thread_pool'
    )    

    def __init__(self, config: Config):
        self.persona_id = str(uuid.uuid4())
        self.logger = HedraLogger()
        self.logger.initialize()

        self.metadata_string: str = None

        self.type = PersonaTypes.DEFAULT
        self.workers = 1

        self.actions = []
        self._hooks: List[Union[ActionHook, TaskHook]] = []
        self.batch = Batch(config)
        self.thread_pool = ThreadPoolExecutor(self.workers)

        self._live_updates = False
        self.total_time = config.total_time
        self.duration = 0
        self.total_actions = 0
        self.total_elapsed = 0
        self.start = 0
        self.end = 0
        self.completed_actions = 0
        self.pending_actions = 0
        self.completed_time = 0
        self.run_timer = False
        self.actions_count = 0
        self.graceful_stop = config.graceful_stop

        self.is_timed = True
        self.timer_thread = None

        self.current_action_idx = 0
        self.optimized_params = None

    def setup(self, hooks: Dict[HookType, List[Union[ActionHook, TaskHook]]], metadata_string: str):

        self.metadata_string = f'{metadata_string} Persona: {self.type.capitalize()}:{self.persona_id} - '

        self._hooks = hooks.get(HookType.ACTION)
        self._hooks.extend(
            hooks.get(HookType.TASK, [])
        )
        self.actions_count = len(self._hooks)
        
            
    async def execute(self):
        hooks = self._hooks
        hook_names = ', '.join([
            hook.name for hook in hooks
        ])

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing {self.actions_count} Hooks: {hook_names}')

        total_time = self.total_time

        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Executing for a total of - {total_time} - seconds')
        loop = asyncio.get_running_loop()

        await self.start_updates()

        self.start = time.monotonic()
        completed, pending = await asyncio.wait([
            loop.create_task(
                hooks[action_idx].session.execute_prepared_request(
                    hooks[action_idx].action
                )
            ) async for action_idx in self.generator(total_time)
        ], timeout=self.graceful_stop)

        self.end = time.monotonic()

        self.pending_actions = len(pending)
        await self.logger.filesystem.aio['hedra.core'].debug(
            f'{self.metadata_string} - Execution completed with - {self.pending_actions} - actions left pending'
        )

        results = await asyncio.gather(*completed)
        
        await self.stop_updates()

        cleanup_start = time.monotonic()

        await asyncio.gather(*[
            asyncio.create_task(
                cancel_pending(pend)
            ) for pend in pending
        ])

        cleanup_elapsed = time.monotonic() - cleanup_start
        await self.logger.filesystem.aio['hedra.core'].info(
            f'{self.metadata_string} - Cleanup completed - Resolved {self.pending_actions} pending actions in {round(cleanup_elapsed, 2)} seconds'
        )

        for hook in hooks:

            session_closed_start = time.monotonic()

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Closing session - {hook.session.session_id} - for Hook - {hook.name}:{hook.hook_id}')
            await hook.session.close()

            session_closed_elapsed = time.monotonic() - session_closed_start

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Closed session - {hook.session.session_id} - for Hook - {hook.name}:{hook.hook_id}. Took: {round(session_closed_elapsed, 2)} seconds')
        
        self.total_actions = len(set(results))
        self.total_elapsed = self.end - self.start
        self.optimized_params = None

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Completed execution')

        return results

    async def generator(self, total_time):
        elapsed = 0
        max_pool_size = int(self.batch.size * (psutil.cpu_count(logical=False) * 2)/self.workers)
        action_idx = 0

        start = time.monotonic()
        while elapsed < total_time:
            yield action_idx
            await asyncio.sleep(0)
            elapsed = time.monotonic() - start

            if self._hooks[action_idx].session.active%max_pool_size == 0:
                try:
                    max_wait = total_time - elapsed
                    await asyncio.wait_for(
                        self._hooks[action_idx].session.wait_for_active_threshold(),
                        timeout=max_wait
                    )
                except asyncio.TimeoutError:
                    pass

            action_idx = (action_idx+1)%self.actions_count

    async def start_updates(self):
        if self._live_updates:
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Live Updates enabled')
            self.timer_thread = asyncio.create_task(self._run_timer_in_background())
    
    async def stop_updates(self):
        if self._live_updates:
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Live Updates stopped')
            self.run_timer = False
            await self.timer_thread

    def run_updates_in_background(self):
        loop = asyncio.new_event_loop()
        loop.run_until_complete(self._run_timer_in_background())

    async def _run_timer_in_background(self):
        start = time.time()
        self.run_timer = True

        while self.run_timer:
            completed_actions = 0
            self.completed_time = time.time() - start

            for deferred_batch in self.batch.deferred:
                completed_actions += await awaitable(len, deferred_batch)


            self.completed_actions = completed_actions
            await asyncio.sleep(1)

        completed_actions = 0
        for deferred_batch in self.batch.deferred:
            completed_actions += await awaitable(len, deferred_batch)

        self.completed_actions = completed_actions
        self.completed_time = time.time() - start
        


# The following code and documentation was inspired, and in some cases
# copied and modified, from the work of Vinay Sajip and contributors
# on cpython's logging package

import abc
import asyncio
import datetime
import enum
import os
import re
import time
from typing import Callable, List, Optional
from aiologger.handlers.base import Handler
from aiologger.records import LogRecord
from aiologger.utils import classproperty, get_running_loop
from hedra.tools.filesystem import open
from hedra.tools.filesystem.text import AsyncTextIOWrapper


class AsyncFileHandler(Handler):
    terminator = "\n"

    def __init__(
        self, filename: str, mode: str = "a", encoding: str = None
    ) -> None:
        super().__init__()
        filename = os.fspath(filename)
        self.absolute_file_path = os.path.abspath(filename)
        self.mode = mode
        self.encoding = encoding
        self.stream: AsyncTextIOWrapper = None
        self._initialization_lock = None

    @property
    def initialized(self):
        return self.stream is not None

    async def _init_writer(self):
        """
        Open the current base file with the (original) mode and encoding.
        """
        if not self._initialization_lock:
            self._initialization_lock = asyncio.Lock()

        async with self._initialization_lock:
            if not self.initialized:
                self.stream = await open(
                    file=self.absolute_file_path,
                    mode=self.mode,
                    encoding=self.encoding,
                )

    async def flush(self):
        await self.stream.flush()

    async def close(self):
        if not self.initialized:
            return
        await self.stream.flush()
        await self.stream.close()
        self.stream = None
        self._initialization_lock = None

    async def emit(self, record: LogRecord):
        if not self.initialized:
            await self._init_writer()

        try:
            msg = self.formatter.format(record)

            # Write order is not guaranteed. String concatenation required
            await self.stream.write(msg + self.terminator)

            await self.stream.flush()
        except Exception as exc:
            await self.handle_error(record, exc)


Namer = Callable[[str], str]
Rotator = Callable[[str, str], None]


class BaseAsyncRotatingFileHandler(AsyncFileHandler, metaclass=abc.ABCMeta):
    def __init__(
        self,
        filename: str,
        mode: str = "a",
        encoding: str = None,
        namer: Namer = None,
        rotator: Rotator = None,
    ) -> None:
        super().__init__(filename, mode, encoding)
        self.mode = mode
        self.encoding = encoding
        self.namer = namer
        self.rotator = rotator
        self._rollover_lock: Optional[asyncio.Lock] = None

    def should_rollover(self, record: LogRecord) -> bool:
        raise NotImplementedError

    async def do_rollover(self):
        raise NotImplementedError

    async def emit(self, record: LogRecord):  # type: ignore
        """
        Emit a record.

        Output the record to the file, catering for rollover as described
        in `do_rollover`.
        """
        try:
            if self.should_rollover(record):
                if not self._rollover_lock:
                    self._rollover_lock = asyncio.Lock()

                async with self._rollover_lock:
                    if self.should_rollover(record):
                        await self.do_rollover()
            await super().emit(record)
        except Exception as exc:
            await self.handle_error(record, exc)

    def rotation_filename(self, default_name: str) -> str:
        """
        Modify the filename of a log file when rotating.

        This is provided so that a custom filename can be provided.

        :param default_name: The default name for the log file.
        """
        if self.namer is None:
            return default_name

        return self.namer(default_name)

    async def rotate(self, source: str, dest: str):
        """
        When rotating, rotate the current log.

        The default implementation calls the 'rotator' attribute of the
        handler, if it's callable, passing the source and dest arguments to
        it. If the attribute isn't callable (the default is None), the source
        is simply renamed to the destination.

        :param source: The source filename. This is normally the base
                       filename, e.g. 'test.log'
        :param dest:   The destination filename. This is normally
                       what the source is rotated to, e.g. 'test.log.1'.
        """
        if self.rotator is None:
            # logging issue 18940: A file may not have been created if delay is True.
            loop = get_running_loop()
            if await loop.run_in_executor(None, lambda: os.path.exists(source)):
                await loop.run_in_executor(  # type: ignore
                    None, lambda: os.rename(source, dest)
                )
        else:
            self.rotator(source, dest)


class RolloverInterval(str, enum.Enum):
    SECONDS = "S"
    MINUTES = "M"
    HOURS = "H"
    DAYS = "D"
    MONDAYS = "W0"
    TUESDAYS = "W1"
    WEDNESDAYS = "W2"
    THURSDAYS = "W3"
    FRIDAYS = "W4"
    SATURDAYS = "W5"
    SUNDAYS = "W6"
    MIDNIGHT = "MIDNIGHT"

    @classproperty
    def WEEK_DAYS(cls):
        return (
            cls.MONDAYS,
            cls.TUESDAYS,
            cls.WEDNESDAYS,
            cls.THURSDAYS,
            cls.FRIDAYS,
            cls.SATURDAYS,
            cls.SUNDAYS,
        )


ONE_MINUTE_IN_SECONDS = 60
ONE_HOUR_IN_SECONDS = 60 * 60
ONE_DAY_IN_SECONDS = ONE_HOUR_IN_SECONDS * 24
ONE_WEEK_IN_SECONDS = 7 * ONE_DAY_IN_SECONDS


class AsyncTimedRotatingFileHandler(BaseAsyncRotatingFileHandler):
    """
    Handler for logging to a file, rotating the log file at certain timed
    intervals.

    If `backup_count` is > 0, when rollover is done, no more than `backup_count`
    files are kept - the oldest ones are deleted.

    | when       | at_time behavior                                       |
    |------------|--------------------------------------------------------|
    | SECONDS    | at_time will be ignored                                |
    | MINUTES    | -- // --                                               |
    | HOURS      | -- // --                                               |
    | DAYS       | at_time will be IGNORED. See also MIDNIGHT             |
    | MONDAYS    | rotation happens every WEEK on MONDAY at ${at_time}    |
    | TUESDAYS   | rotation happens every WEEK on TUESDAY at ${at_time}   |
    | WEDNESDAYS | rotation happens every WEEK on WEDNESDAY at ${at_time} |
    | THURSDAYS  | rotation happens every WEEK on THURSDAY at ${at_time}  |
    | FRIDAYS    | rotation happens every WEEK on FRIDAY at ${at_time}    |
    | SATURDAYS  | rotation happens every WEEK on SATURDAY at ${at_time}  |
    | SUNDAYS    | rotation happens every WEEK on SUNDAY at ${at_time}    |
    | MIDNIGHT   | rotation happens every DAY at ${at_time}               |
    """

    def __init__(
        self,
        filename: str,
        when: RolloverInterval = RolloverInterval.HOURS,
        interval: int = 1,
        backup_count: int = 0,
        encoding: str = None,
        utc: bool = False,
        at_time: datetime.time = None,
    ) -> None:
        super().__init__(filename=filename, mode="a", encoding=encoding)
        self.when = when.upper()
        self.backup_count = backup_count
        self.utc = utc
        self.at_time = at_time
        # Calculate the real rollover interval, which is just the number of
        # seconds between rollovers.  Also set the filename suffix used when
        # a rollover occurs.  Current 'when' events supported:
        # S - Seconds
        # M - Minutes
        # H - Hours
        # D - Days
        # midnight - roll over at midnight
        # W{0-6} - roll over on a certain day; 0 - Monday
        #
        # Case of the 'when' specifier is not important; lower or upper case
        # will work.
        if self.when == RolloverInterval.SECONDS:
            self.interval = 1  # one second
            self.suffix = "%Y-%m-%d_%H-%M-%S"
            ext_match = r"^\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}(\.\w+)?$"
        elif self.when == RolloverInterval.MINUTES:
            self.interval = ONE_MINUTE_IN_SECONDS  # one minute
            self.suffix = "%Y-%m-%d_%H-%M"
            ext_match = r"^\d{4}-\d{2}-\d{2}_\d{2}-\d{2}(\.\w+)?$"
        elif self.when == RolloverInterval.HOURS:
            self.interval = ONE_HOUR_IN_SECONDS  # one hour
            self.suffix = "%Y-%m-%d_%H"
            ext_match = r"^\d{4}-\d{2}-\d{2}_\d{2}(\.\w+)?$"
        elif (
            self.when == RolloverInterval.DAYS
            or self.when == RolloverInterval.MIDNIGHT
        ):
            self.interval = ONE_DAY_IN_SECONDS  # one day
            self.suffix = "%Y-%m-%d"
            ext_match = r"^\d{4}-\d{2}-\d{2}(\.\w+)?$"
        elif self.when.startswith("W"):
            if self.when not in RolloverInterval.WEEK_DAYS:
                raise ValueError(
                    f"Invalid day specified for weekly rollover: {self.when}"
                )
            self.interval = ONE_DAY_IN_SECONDS * 7  # one week
            self.day_of_week = int(self.when[1])
            self.suffix = "%Y-%m-%d"
            ext_match = r"^\d{4}-\d{2}-\d{2}(\.\w+)?$"
        else:
            raise ValueError(f"Invalid RolloverInterval specified: {self.when}")

        self.ext_match = re.compile(ext_match, re.ASCII)
        self.interval = self.interval * interval  # multiply by units requested
        # The following line added because the filename passed in could be a
        # path object (see Issue #27493), but self.baseFilename will be a string
        filename = self.absolute_file_path
        if os.path.exists(filename):  # todo: IO. Remove or postpone
            t = int(os.stat(filename).st_mtime)
        else:
            t = int(time.time())
        self.rollover_at = self.compute_rollover(t)

    def compute_rollover(self, current_time: int) -> int:
        """
        Work out the rollover time based on the specified time.

        If we are rolling over at midnight or weekly, then the interval is
        already known. need to figure out is WHEN the next interval is.
        In other words, if you are rolling over at midnight, then your base
        interval is 1 day, but you want to start that one day clock at midnight,
         not now. So, we have to fudge the `rollover_at` value in order to trigger
         the first rollover at the right time.  After that, the regular interval
         will take care of the rest.  Note that this code doesn't care about
         leap seconds. :)
        """
        result = current_time + self.interval

        if (
            self.when == RolloverInterval.MIDNIGHT
            or self.when in RolloverInterval.WEEK_DAYS
        ):
            if self.utc:
                t = time.gmtime(current_time)
            else:
                t = time.localtime(current_time)
            current_hour = t[3]
            current_minute = t[4]
            current_second = t[5]
            current_day = t[6]
            # r is the number of seconds left between now and the next rotation
            if self.at_time is None:
                rotate_ts = ONE_DAY_IN_SECONDS
            else:
                rotate_ts = (
                    self.at_time.hour * 60 + self.at_time.minute
                ) * 60 + self.at_time.second

            r = rotate_ts - (
                (current_hour * 60 + current_minute) * 60 + current_second
            )
            if r < 0:
                # Rotate time is before the current time (for example when
                # self.rotateAt is 13:45 and it now 14:15), rotation is
                # tomorrow.
                r += ONE_DAY_IN_SECONDS
                current_day = (current_day + 1) % 7
            result = current_time + r
            # If we are rolling over on a certain day, add in the number of days until
            # the next rollover, but offset by 1 since we just calculated the time
            # until the next day starts.  There are three cases:
            # Case 1) The day to rollover is today; in this case, do nothing
            # Case 2) The day to rollover is further in the interval (i.e., today is
            #         day 2 (Wednesday) and rollover is on day 6 (Sunday).  Days to
            #         next rollover is simply 6 - 2 - 1, or 3.
            # Case 3) The day to rollover is behind us in the interval (i.e., today
            #         is day 5 (Saturday) and rollover is on day 3 (Thursday).
            #         Days to rollover is 6 - 5 + 3, or 4.  In this case, it's the
            #         number of days left in the current week (1) plus the number
            #         of days in the next week until the rollover day (3).
            # The calculations described in 2) and 3) above need to have a day added.
            # This is because the above time calculation takes us to midnight on this
            # day, i.e. the start of the next day.
            if self.when in RolloverInterval.WEEK_DAYS:
                day = current_day  # 0 is Monday
                if day != self.day_of_week:
                    if day < self.day_of_week:
                        days_to_wait = self.day_of_week - day
                    else:
                        days_to_wait = 6 - day + self.day_of_week + 1
                    new_rollover_at = result + (
                        days_to_wait * ONE_DAY_IN_SECONDS
                    )
                    if not self.utc:
                        dst_now = t[-1]
                        dst_at_rollover = time.localtime(new_rollover_at)[-1]
                        if dst_now != dst_at_rollover:
                            if not dst_now:
                                # DST kicks in before next rollover, so we need to deduct an hour
                                new_rollover_at -= ONE_HOUR_IN_SECONDS
                            else:
                                # DST bows out before next rollover, so we need to add an hour
                                new_rollover_at += ONE_HOUR_IN_SECONDS
                    result = new_rollover_at
        return result

    def should_rollover(self, record: LogRecord) -> bool:
        """
        Determine if rollover should occur.

        record is not used, as we are just comparing times, but it is needed so
        the method signatures are the same
        """
        t = int(time.time())
        if t >= self.rollover_at:
            return True
        return False

    async def get_files_to_delete(self) -> List[str]:
        """
        Determine the files to delete when rolling over.
        """
        dir_name, base_name = os.path.split(self.absolute_file_path)
        loop = get_running_loop()
        file_names = await loop.run_in_executor(
            None, lambda: os.listdir(dir_name)
        )
        result = []
        prefix = base_name + "."
        plen = len(prefix)
        for file_name in file_names:
            if file_name[:plen] == prefix:
                suffix = file_name[plen:]
                if self.ext_match.match(suffix):
                    result.append(os.path.join(dir_name, file_name))
        if len(result) < self.backup_count:
            return []
        else:
            result.sort(reverse=True)  # os.listdir order is not defined
            return result[: len(result) - self.backup_count]

    async def _delete_files(self, file_paths: List[str]):
        loop = get_running_loop()
        for file_path in file_paths:
            await loop.run_in_executor(  # type: ignore
                None, lambda: os.unlink(file_path)
            )

    async def do_rollover(self):
        """
        do a rollover; in this case, a date/time stamp is appended to the filename
        when the rollover happens.  However, you want the file to be named for the
        start of the interval, not the current time.  If there is a backup count,
        then we have to get a list of matching filenames, sort them and remove
        the one with the oldest suffix.
        """
        if self.stream:
            await self.stream.close()
            self.stream = None
        # get the time that this sequence started at and make it a TimeTuple
        current_time = int(time.time())
        dst_now = time.localtime(current_time)[-1]
        t = self.rollover_at - self.interval
        if self.utc:
            time_tuple = time.gmtime(t)
        else:
            time_tuple = time.localtime(t)
            dst_then = time_tuple[-1]
            if dst_now != dst_then:
                if dst_now:
                    addend = ONE_HOUR_IN_SECONDS
                else:
                    addend = -ONE_HOUR_IN_SECONDS
                time_tuple = time.localtime(t + addend)
        destination_file_path = self.rotation_filename(
            self.absolute_file_path
            + "."
            + time.strftime(self.suffix, time_tuple)
        )
        loop = get_running_loop()
        if await loop.run_in_executor(
            None, lambda: os.path.exists(destination_file_path)
        ):
            await loop.run_in_executor(
                None, lambda: os.unlink(destination_file_path)
            )
        await self.rotate(self.absolute_file_path, destination_file_path)
        if self.backup_count > 0:
            files_to_delete = await self.get_files_to_delete()
            if files_to_delete:
                await self._delete_files(files_to_delete)

        await self._init_writer()
        new_rollover_at = self.compute_rollover(current_time)
        while new_rollover_at <= current_time:
            new_rollover_at = new_rollover_at + self.interval
        # If DST changes and midnight or weekly rollover, adjust for this.
        if (
            self.when == RolloverInterval.MIDNIGHT
            or self.when in RolloverInterval.WEEK_DAYS
        ) and not self.utc:
            dst_at_rollover = time.localtime(new_rollover_at)[-1]
            if dst_now != dst_at_rollover:
                if not dst_now:
                    # DST kicks in before next rollover, so we need to deduct an hour
                    addend = -ONE_HOUR_IN_SECONDS
                else:
                    # DST bows out before next rollover, so we need to add an hour
                    addend = ONE_HOUR_IN_SECONDS
                new_rollover_at += addend
        self.rollover_at = new_rollover_at



class ServiceNode:

    def __init__(self, service_config) -> None:
        self.service_address = service_config.get('serviceAddress')
        self.service_id = service_config.get('serviceId')
        self.service_server_address = service_config.get('serviceServerAddress')
        self.service_name = service_config.get('serviceName')



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: worker.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cworker.proto\x1a\x1cgoogle/protobuf/struct.proto\">\n\x13LeaderUpdateRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x11\n\thost_port\x18\x02 \x01(\x03\"u\n\x1aWorkerServerUpdateResponse\x12\x16\n\x0e\x63lient_address\x18\x01 \x01(\t\x12\x13\n\x0b\x63lient_port\x18\x02 \x01(\x03\x12*\n\tcompleted\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\"\x88\x01\n\rNewJobRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x0e\n\x06job_id\x18\x03 \x01(\t\x12+\n\njob_config\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x13\n\x0bjob_timeout\x18\x05 \x01(\x03\"G\n\x0eNewJobResponse\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x0e\n\x06job_id\x18\x03 \x01(\t\"J\n\x11PollWorkerRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x0e\n\x06job_id\x18\x03 \x01(\t\"N\n\x12PollWorkerResponse\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x11\n\tcompleted\x18\x03 \x01(\x08\"K\n\x12JobCompleteRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x0e\n\x06job_id\x18\x03 \x01(\t\"j\n\x13JobCompleteResponse\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12,\n\x0bjob_results\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\"O\n\x16WorkerHeartbeatRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x0e\n\x06status\x18\x03 \x01(\t\"P\n\x17WorkerHeartbeatResponse\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x0e\n\x06status\x18\x03 \x01(\t\"D\n\x1bLeaderRegisterWorkerRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\"]\n\x1cLeaderRegisterWorkerResponse\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x16\n\x0eserver_address\x18\x03 \x01(\t2\xaf\x03\n\x0cWorkerServer\x12K\n\x14GetRequestsCompleted\x12\x14.LeaderUpdateRequest\x1a\x1b.WorkerServerUpdateResponse\"\x00\x12\x37\n\x12\x43reateNewWorkerJob\x12\x0e.NewJobRequest\x1a\x0f.NewJobResponse\"\x00\x12\x39\n\x0cGetJobStatus\x12\x12.PollWorkerRequest\x1a\x13.PollWorkerResponse\"\x00\x12<\n\rGetJobResults\x12\x13.JobCompleteRequest\x1a\x14.JobCompleteResponse\"\x00\x12K\n\x14\x43heckWorkerHeartbeat\x12\x17.WorkerHeartbeatRequest\x1a\x18.WorkerHeartbeatResponse\"\x00\x12S\n\x14LeaderRegisterWorker\x12\x1c.LeaderRegisterWorkerRequest\x1a\x1d.LeaderRegisterWorkerResponseb\x06proto3')



_LEADERUPDATEREQUEST = DESCRIPTOR.message_types_by_name['LeaderUpdateRequest']
_WORKERSERVERUPDATERESPONSE = DESCRIPTOR.message_types_by_name['WorkerServerUpdateResponse']
_NEWJOBREQUEST = DESCRIPTOR.message_types_by_name['NewJobRequest']
_NEWJOBRESPONSE = DESCRIPTOR.message_types_by_name['NewJobResponse']
_POLLWORKERREQUEST = DESCRIPTOR.message_types_by_name['PollWorkerRequest']
_POLLWORKERRESPONSE = DESCRIPTOR.message_types_by_name['PollWorkerResponse']
_JOBCOMPLETEREQUEST = DESCRIPTOR.message_types_by_name['JobCompleteRequest']
_JOBCOMPLETERESPONSE = DESCRIPTOR.message_types_by_name['JobCompleteResponse']
_WORKERHEARTBEATREQUEST = DESCRIPTOR.message_types_by_name['WorkerHeartbeatRequest']
_WORKERHEARTBEATRESPONSE = DESCRIPTOR.message_types_by_name['WorkerHeartbeatResponse']
_LEADERREGISTERWORKERREQUEST = DESCRIPTOR.message_types_by_name['LeaderRegisterWorkerRequest']
_LEADERREGISTERWORKERRESPONSE = DESCRIPTOR.message_types_by_name['LeaderRegisterWorkerResponse']
LeaderUpdateRequest = _reflection.GeneratedProtocolMessageType('LeaderUpdateRequest', (_message.Message,), {
  'DESCRIPTOR' : _LEADERUPDATEREQUEST,
  '__module__' : 'worker_pb2'
  # @@protoc_insertion_point(class_scope:LeaderUpdateRequest)
  })
_sym_db.RegisterMessage(LeaderUpdateRequest)

WorkerServerUpdateResponse = _reflection.GeneratedProtocolMessageType('WorkerServerUpdateResponse', (_message.Message,), {
  'DESCRIPTOR' : _WORKERSERVERUPDATERESPONSE,
  '__module__' : 'worker_pb2'
  # @@protoc_insertion_point(class_scope:WorkerServerUpdateResponse)
  })
_sym_db.RegisterMessage(WorkerServerUpdateResponse)

NewJobRequest = _reflection.GeneratedProtocolMessageType('NewJobRequest', (_message.Message,), {
  'DESCRIPTOR' : _NEWJOBREQUEST,
  '__module__' : 'worker_pb2'
  # @@protoc_insertion_point(class_scope:NewJobRequest)
  })
_sym_db.RegisterMessage(NewJobRequest)

NewJobResponse = _reflection.GeneratedProtocolMessageType('NewJobResponse', (_message.Message,), {
  'DESCRIPTOR' : _NEWJOBRESPONSE,
  '__module__' : 'worker_pb2'
  # @@protoc_insertion_point(class_scope:NewJobResponse)
  })
_sym_db.RegisterMessage(NewJobResponse)

PollWorkerRequest = _reflection.GeneratedProtocolMessageType('PollWorkerRequest', (_message.Message,), {
  'DESCRIPTOR' : _POLLWORKERREQUEST,
  '__module__' : 'worker_pb2'
  # @@protoc_insertion_point(class_scope:PollWorkerRequest)
  })
_sym_db.RegisterMessage(PollWorkerRequest)

PollWorkerResponse = _reflection.GeneratedProtocolMessageType('PollWorkerResponse', (_message.Message,), {
  'DESCRIPTOR' : _POLLWORKERRESPONSE,
  '__module__' : 'worker_pb2'
  # @@protoc_insertion_point(class_scope:PollWorkerResponse)
  })
_sym_db.RegisterMessage(PollWorkerResponse)

JobCompleteRequest = _reflection.GeneratedProtocolMessageType('JobCompleteRequest', (_message.Message,), {
  'DESCRIPTOR' : _JOBCOMPLETEREQUEST,
  '__module__' : 'worker_pb2'
  # @@protoc_insertion_point(class_scope:JobCompleteRequest)
  })
_sym_db.RegisterMessage(JobCompleteRequest)

JobCompleteResponse = _reflection.GeneratedProtocolMessageType('JobCompleteResponse', (_message.Message,), {
  'DESCRIPTOR' : _JOBCOMPLETERESPONSE,
  '__module__' : 'worker_pb2'
  # @@protoc_insertion_point(class_scope:JobCompleteResponse)
  })
_sym_db.RegisterMessage(JobCompleteResponse)

WorkerHeartbeatRequest = _reflection.GeneratedProtocolMessageType('WorkerHeartbeatRequest', (_message.Message,), {
  'DESCRIPTOR' : _WORKERHEARTBEATREQUEST,
  '__module__' : 'worker_pb2'
  # @@protoc_insertion_point(class_scope:WorkerHeartbeatRequest)
  })
_sym_db.RegisterMessage(WorkerHeartbeatRequest)

WorkerHeartbeatResponse = _reflection.GeneratedProtocolMessageType('WorkerHeartbeatResponse', (_message.Message,), {
  'DESCRIPTOR' : _WORKERHEARTBEATRESPONSE,
  '__module__' : 'worker_pb2'
  # @@protoc_insertion_point(class_scope:WorkerHeartbeatResponse)
  })
_sym_db.RegisterMessage(WorkerHeartbeatResponse)

LeaderRegisterWorkerRequest = _reflection.GeneratedProtocolMessageType('LeaderRegisterWorkerRequest', (_message.Message,), {
  'DESCRIPTOR' : _LEADERREGISTERWORKERREQUEST,
  '__module__' : 'worker_pb2'
  # @@protoc_insertion_point(class_scope:LeaderRegisterWorkerRequest)
  })
_sym_db.RegisterMessage(LeaderRegisterWorkerRequest)

LeaderRegisterWorkerResponse = _reflection.GeneratedProtocolMessageType('LeaderRegisterWorkerResponse', (_message.Message,), {
  'DESCRIPTOR' : _LEADERREGISTERWORKERRESPONSE,
  '__module__' : 'worker_pb2'
  # @@protoc_insertion_point(class_scope:LeaderRegisterWorkerResponse)
  })
_sym_db.RegisterMessage(LeaderRegisterWorkerResponse)

_WORKERSERVER = DESCRIPTOR.services_by_name['WorkerServer']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _LEADERUPDATEREQUEST._serialized_start=46
  _LEADERUPDATEREQUEST._serialized_end=108
  _WORKERSERVERUPDATERESPONSE._serialized_start=110
  _WORKERSERVERUPDATERESPONSE._serialized_end=227
  _NEWJOBREQUEST._serialized_start=230
  _NEWJOBREQUEST._serialized_end=366
  _NEWJOBRESPONSE._serialized_start=368
  _NEWJOBRESPONSE._serialized_end=439
  _POLLWORKERREQUEST._serialized_start=441
  _POLLWORKERREQUEST._serialized_end=515
  _POLLWORKERRESPONSE._serialized_start=517
  _POLLWORKERRESPONSE._serialized_end=595
  _JOBCOMPLETEREQUEST._serialized_start=597
  _JOBCOMPLETEREQUEST._serialized_end=672
  _JOBCOMPLETERESPONSE._serialized_start=674
  _JOBCOMPLETERESPONSE._serialized_end=780
  _WORKERHEARTBEATREQUEST._serialized_start=782
  _WORKERHEARTBEATREQUEST._serialized_end=861
  _WORKERHEARTBEATRESPONSE._serialized_start=863
  _WORKERHEARTBEATRESPONSE._serialized_end=943
  _LEADERREGISTERWORKERREQUEST._serialized_start=945
  _LEADERREGISTERWORKERREQUEST._serialized_end=1013
  _LEADERREGISTERWORKERRESPONSE._serialized_start=1015
  _LEADERREGISTERWORKERRESPONSE._serialized_end=1108
  _WORKERSERVER._serialized_start=1111
  _WORKERSERVER._serialized_end=1542
# @@protoc_insertion_point(module_scope)



from .data_received_event import DataReceived
from .headers_sent_event import HeadersSent
from .informational_respose_received_event import InformationalResponseReceived
from .remote_settings_changed_event import RemoteSettingsChanged
from .response_received_event import ResponseReceived
from .response_sent_event import ResponseSent
from .request_received_event import RequestReceived
from .request_sent_event import RequestSent
from .settings_acknowledged_event import SettingsAcknowledged
from .stream_ended_event import StreamEnded
from .stream_reset import StreamReset
from .trailers_sent_event import TrailersSent
from .trailers_received_event import TrailersReceived
from .window_updated_event import WindowUpdated
from .deferred_headers_event import DeferredHeaders
from .connection_terminated_event import ConnectionTerminated


import asyncio
import psutil
import traceback
from typing_extensions import TypeVarTuple, Unpack
from typing import Dict, Generic, List, Any, Union, Coroutine
from hedra.core.graphs.events import Event
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.hooks.hook_types.internal import Internal
from hedra.core.engines.client.client import Client
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.common.base_action import BaseAction
from hedra.core.engines.types.task.task import Task
from hedra.core.graphs.hooks.registry.registry_types import (
    ActionHook,
    AfterHook,
    BeforeHook,
    ChannelHook,
    CheckHook,
    ContextHook,
    EventHook,
    SetupHook,
    TaskHook
)
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.personas.types import PersonaTypesMap
from hedra.logging import HedraLogger
from hedra.plugins.types.engine.engine_plugin import EnginePlugin
from hedra.plugins.types.plugin_types import PluginType
from playwright.async_api import Geolocation
from hedra.core.graphs.stages.execute import Execute
from hedra.core.graphs.stages.base.stage import Stage
from .exceptions import HookSetupError

T = TypeVarTuple('T')


class SetupCall:

    def __init__(self, hook: Hook, retries: int=1) -> None:
        self.hook = hook
        self.hook_name = self.hook.hook_type.name.capitalize()
        self.exception = None
        self.action_store = None
        self.retries = retries
        self.current_try = 1
        self.logger = HedraLogger()
        self.logger.initialize()
        self.metadata_string:str = None
        self.error_traceback: str = None

    async def setup(self):

        for _ in range(self.retries):
            try:
                await self.hook.call()

            except Exception as setup_exception:
                self.exception = setup_exception

                self.error_traceback = str(traceback.format_exc())
                await self.logger.spinner.system.error(f'{self.metadata_string} - Encountered connection validation error - {str(setup_exception)} - {self.hook_name} Hook: {self.hook.name}:{self.hook.hook_id}')
                await self.logger.filesystem.aio['hedra.core'].error(f'{self.metadata_string} - Encountered connection validation error - {str(setup_exception)} - {self.hook_name} Hook: {self.hook.name}:{self.hook.hook_id}')

                if self.current_try >= self.retries:
                    self.action_store.waiter.set_result(None)
                    break

                else:
                    self.current_try += 1



class Setup(Stage, Generic[Unpack[T]]):
    stage_type=StageTypes.SETUP
    log_level='info'
    persona_type='default'
    total_time='1m'
    batch_size=1000
    batch_interval=0
    action_interval=0
    batch_gradient=0.1
    cpus=int(psutil.cpu_count(logical=False))
    no_run_visuals=False
    graceful_stop=1
    connect_timeout=10
    request_timeout=60
    reset_connections=False
    apply_to_stages=[]
    browser_type: str='chromium'
    device_type: str=None
    locale: str=None
    geolocation: Geolocation=None
    permissions: List[str]=[]
    playwright_options: Dict[str, Any]={}

    
    def __init__(self) -> None:
        super().__init__()
        self.generation_setup_candidates = 0
        self.stages: Dict[str, Execute] = {}
        self.accepted_hook_types = [ HookType.SETUP, HookType.EVENT, HookType.CONTEXT ]
        self.persona_types = PersonaTypesMap()
        self.config = Config(
            log_level=self.log_level,
            persona_type=self.persona_types[self.persona_type],
            total_time=self.total_time,
            batch_size=self.batch_size,
            batch_interval=self.batch_interval,
            action_interval=self.action_interval,
            batch_gradient=self.batch_gradient,
            cpus=self.cpus,
            no_run_visuals=self.no_run_visuals,
            connect_timeout=self.connect_timeout,
            request_timeout=self.request_timeout,
            graceful_stop=self.graceful_stop,
            reset_connections=self.reset_connections,
            browser_type=self.browser_type,
            device_type=self.device_type,
            locale=self.locale,
            geolocation=self.geolocation,
            permissions=self.permissions,
            playwright_options=self.playwright_options
        )

        self.client = Client(
            self.graph_name,
            self.graph_id,
            self.name,
            self.stage_id
        )
        self.client._config = self.config

        self.internal_hooks.extend([
            'get_hook',
            'get_checks',
            'setup'
        ])

    @Internal()
    async def run(self):

        events: List[Union[EventHook, Event]] = [event for event in self.hooks[HookType.EVENT]]
        pre_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre
        ]
        
        if len(pre_events) > 0:
            pre_event_names = ", ".join([
                event.shortname for event in pre_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing PRE events - {pre_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in pre_events
            ], timeout=self.stage_timeout)
        
        bypass_connection_validation = self.core_config.get('bypass_connection_validation', False)
        connection_validation_retries = self.core_config.get('connection_validation_retries', 3)

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Starting setup')

        setup_hooks: List[SetupHook] = self.hooks[HookType.SETUP]
        setup_hook_names = ', '.join([hook.name for hook in setup_hooks])
        
        if len(setup_hooks) > 0:
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Runnning Setup hooks for stage - {setup_hook_names}')
        
        await asyncio.gather(*[hook.call() for hook in setup_hooks])
        execute_stage_id = 1

        stages = dict(self.stages)

        execute_stage_names = ', '.join(list(stages.keys()))

        await self.logger.spinner.append_message(f'Setting up - {execute_stage_names}')
        
        for execute_stage_name, execute_stage in stages.items():

            execute_stage.execution_stage_id = execute_stage_id
            execute_stage.execute_setup_stage = self.name

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Execute stage - {execute_stage_name} - assigned stage order id - {execute_stage_id}')

            execute_stage_id += 1

            persona_plugins = self.plugins_by_type.get(PluginType.PERSONA)
            for plugin_name in persona_plugins.keys():
                self.persona_types.types[plugin_name] = plugin_name
                await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Loaded Persona plugin - {plugin.name} - for Execute stae - {execute_stage_name}')
   
            client = Client(
                self.graph_name,
                self.graph_id,
                execute_stage.name,
                execute_stage.stage_id
            )
            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Created Client, id - {client.client_id} - for Execute stage - {execute_stage_name}')

            engine_plugins: Dict[str, EnginePlugin] = self.plugins_by_type.get(PluginType.ENGINE)

            for plugin_name, plugin in engine_plugins.items():
                client.plugin[plugin_name] = plugin(self.config)
                plugin.name = plugin_name
                self.plugins_by_type[plugin_name] = plugin

                await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Loaded Engine plugin - {plugin.name} - for Execute stage - {execute_stage_name}')


            execute_stage.client = client
            execute_stage.client._config = self.config

            action_hooks: List[ActionHook] = execute_stage.hooks[HookType.ACTION]
            for hook in action_hooks:

                execute_stage.client.next_name = hook.name
                execute_stage.client.intercept = True

                await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Client intercept set to {execute_stage.client.intercept} - Action calls for client id - {execute_stage.client.client_id} - will be suspended on execution')

                await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Setting up Action - {hook.name}:{hook.hook_id} - for Execute stage - {execute_stage_name}')
                await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Preparing Action hook - {hook.name}:{hook.hook_id} - for suspension - Execute stage - {execute_stage_name}')

                execute_stage.client.actions.set_waiter(execute_stage.name)

                setup_call = SetupCall(hook, retries=connection_validation_retries)

                setup_call.metadata_string = self.metadata_string
                setup_call.action_store = execute_stage.client.actions

                await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Executing Action hook call - {hook.name}:{hook.hook_id} - Execute stage - {execute_stage_name}')

                task = asyncio.create_task(setup_call.setup())
                await execute_stage.client.actions.wait_for_ready(setup_call)   

                await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Exiting suspension for Action - {hook.name}:{hook.hook_id} - Execute stage - {execute_stage_name}')

                action = None
                session = None

                try:
                    if setup_call.exception:
                        raise HookSetupError(hook, HookType.ACTION, str(setup_call.exception))

                    task.cancel()
                    if task.cancelled() is False:
                        await asyncio.wait_for(task, timeout=0.1)

                except HookSetupError as hook_setup_exception:

                    if bypass_connection_validation:

                        hook.hook_type = HookType.TASK

                        execute_stage.hooks[HookType.TASK].append(hook)
                        action_idx = execute_stage.hooks[HookType.ACTION].index(hook)
                        execute_stage.hooks[HookType.ACTION].pop(action_idx)

                    else:
                        raise hook_setup_exception

                except asyncio.InvalidStateError:
                    pass

                except asyncio.CancelledError:
                    pass

                except asyncio.TimeoutError:
                    pass

                if hook.hook_type == HookType.ACTION:

                    action, session = execute_stage.client.actions.get(
                        execute_stage.name,
                        hook.name
                    )

                    await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Successfully retrieved prepared Action and Session for action - {action.name}:{action.action_id} - Execute stage - {execute_stage_name}')
                    
                    action.hooks.before =  await self.get_hook(execute_stage, hook.shortname, HookType.BEFORE)
                    action.hooks.after = await self.get_hook(execute_stage, hook.shortname, HookType.AFTER)
                    action.hooks.checks = await self.get_checks(execute_stage, hook.shortname)
                    
                    hook.session = session
                    hook.action = action    

            task_hooks: List[TaskHook] = execute_stage.hooks[HookType.TASK]
            for hook in task_hooks:

                await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Loading Task hook - {hook.name}:{hook.hook_id} - to Execute stage - {execute_stage_name}')

                execute_stage.client.next_name = hook.name
                task, session = execute_stage.client.task.call(
                    hook.call,
                    env=hook.metadata.env,
                    user=hook.metadata.user,
                    tags=hook.metadata.tags
                )

                await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Successfully retrieved task and session for Task - {task.name}:{task.action_id} - Execute stage - {execute_stage_name}')

                task.hooks.checks = await self.get_checks(execute_stage, hook.shortname)

                hook.session = session
                hook.action = task  

            await self.get_channels(execute_stage)

            execute_stage.client.intercept = False
            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Client intercept set to {execute_stage.client.intercept} - Action calls for client id - {execute_stage.client.client_id} - will not be suspended on execution')

            for setup_hook in execute_stage.hooks[HookType.SETUP]:
                await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing Setup hook - {setup_hook.name} - for Execute stage - {execute_stage_name}')

                await setup_hook.call()

            self.stages[execute_stage_name] = execute_stage

            actions_generated_count = len(execute_stage.hooks[HookType.ACTION])
            tasks_generated_count = len(execute_stage.hooks[HookType.TASK])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Generated - {actions_generated_count} - Actions for Execute stage - {execute_stage_name}')
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Generated - {tasks_generated_count} - Tasks for Execute stage - {execute_stage_name}')

        post_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre is False
        ]

        if len(post_events) > 0:
            post_event_names = ", ".join([
                event.shortname for event in post_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing POST events - {post_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in post_events
            ], timeout=self.stage_timeout)


        await self.logger.spinner.set_default_message(f'Setup for - {execute_stage_names} - complete')
        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Completed setup')

        context_hooks: List[ContextHook] = self.hooks[HookType.CONTEXT]
        context_hooks: List[ContextHook] = self.hooks[HookType.CONTEXT]
        await asyncio.gather(*[
            asyncio.create_task(context_hook.call(self.context)) for context_hook in context_hooks
        ])

        return self.stages

    @Internal()
    async def get_hook(self, execute_stage: Execute, shortname: str, hook_type: str) -> Coroutine:
        hooks: List[Union[ActionHook, BeforeHook]] = execute_stage.hooks[hook_type]
        for hook in hooks:
            if shortname in hook.names:
                await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Adding Hook - {hook.name}:{hook.hook_id} - of type - {hook.hook_type.name.capitalize()} - to Action - {shortname} - for Execute stage - {execute_stage.name}')

                return hook.call

    @Internal()
    async def get_checks(self, execute_stage: Execute, shortname: str) -> List[Coroutine]:

        checks = []
        checks_hooks: List[CheckHook] = execute_stage.hooks[HookType.CHECK]
        for hook in checks_hooks:
            if shortname in hook.names:
                await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Adding Check - {hook.name}:{hook.hook_id} - to Action or Task - {shortname} - for Execute stage - {execute_stage.name}')
                
                checks.append(hook.call)

        return checks

    @Internal()
    async def get_channels(self, execute_stage: Execute) -> None:
        listeners: Dict[str, Union[ActionHook, TaskHook]] = {}
        notifiers: Dict[str, Union[ActionHook, TaskHook]] = {}
        channels: Dict[str, ChannelHook] = {channel.shortname: channel for channel in execute_stage.hooks[HookType.CHANNEL]}

        actions_and_tasks: List[Union[ActionHook, TaskHook]] = execute_stage.hooks[HookType.ACTION]
        actions_and_tasks.extend(execute_stage.hooks[HookType.TASK])

        for hook in actions_and_tasks:
            action: Union[BaseAction, Task] = hook.action
            if hook.is_listener:
                listeners[hook.name] = hook

                for channel_name in hook.listeners:
                    channel: ChannelHook = channels.get(channel_name)
                    action.hooks.channels.append(channel.call)
                    channel.listeners.append(hook.name)

            if hook.is_notifier:
                notifiers[hook.name] = hook

                for channel_name in hook.notifiers:
                    channel: ChannelHook = channels.get(channel_name)
                    action.hooks.channels.append(channel.call)
                    channel.notifiers.append(hook.name)

        for channel in channels.values():
            for notifier_name in channel.notifiers:
                notifier: Union[ActionHook, TaskHook] = notifiers.get(notifier_name)
                action: Union[BaseAction, Task] = notifier.action
                
                action.hooks.listeners: List[Hook] = [listeners.get(listener_name) for listener_name in channel.listeners]

    @Internal()
    async def setup(self) -> None:
        setup_hooks: List[SetupHook] = self.hooks[HookType.SETUP]
        for setup_hook in setup_hooks:
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing Setup hook - {setup_hook.name}')

            await setup_hook()


import os
import json
import json
from urllib.parse import urlparse
from hedra.projects.management import GraphManager
from hedra.projects.management.graphs.actions import RepoConfig
from hedra.cli.exceptions.graph.sync import NotSetError
from hedra.logging import HedraLogger
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)


def get_project(
    url: str, 
    path: str,
    branch: str, 
    remote: str, 
    username: str, 
    password: str,
    bypass_connection_validation: bool,
    connection_validation_retries: int,
    log_level: str
):
    logging_manager.disable(
        LoggerTypes.HEDRA, 
        LoggerTypes.DISTRIBUTED,
        LoggerTypes.FILESYSTEM,
        LoggerTypes.DISTRIBUTED_FILESYSTEM
    )

    logging_manager.update_log_level(log_level)

    logger = HedraLogger()
    logger.initialize()
    logger['console'].sync.info(f'Fetching project at - {url} - and saving at - {path}...')

    hedra_config_filepath = os.path.join(
        path,
        '.hedra.json'
    )

    hedra_config = {}
    if os.path.exists(hedra_config_filepath):
        with open(hedra_config_filepath, 'r') as hedra_config_file:
            hedra_config = json.load(hedra_config_file)

    hedra_project_config = hedra_config.get('project', {})


    if os.path.exists(path) is False:
        os.mkdir(path)

    if url is None:
        url = hedra_project_config.get('project_url')
    
    if username is None:
        username = hedra_project_config.get('project_username')

    if password is None:
        password = hedra_project_config.get('project_password')

    if branch is None:
        branch = hedra_project_config.get('project_branch', 'main')

    if remote is None:
        remote = hedra_project_config.get('project_remote', 'origin')

    if url is None:
        raise NotSetError(
            path,
            'url',
            '--url'
        )

    if username is None:
        raise NotSetError(
            path,
            'username',
            '--username'
        )

    if password is None:
        raise NotSetError(
            path,
            'password',
            '--password'
        )

    parsed_url = urlparse(url)
    repo_url = f'{parsed_url.scheme}://{username}:{password}@{parsed_url.hostname}{parsed_url.path}'

    logger['console'].sync.info('Initializing project manager.')

    repo_config = RepoConfig(
        path,
        repo_url,
        branch=branch,
        remote=remote,
        username=username,
        password=password
    )
    
    manager = GraphManager(repo_config, log_level=log_level)
    workflow_actions = [
            'fetch'
        ]

    manager.execute_workflow(workflow_actions)

    hedra_project_config.update({
        'project_url': url,
        'project_username': username,
        'project_password': password,
        'project_branch': branch,
        'project_remote': remote
    })

    discovered = manager.discover_graph_files()
    hedra_config.update(discovered)

    new_graphs_count = len({
        graph_name for graph_name in discovered['graphs'] if graph_name not in hedra_config['graphs']
    })

    new_plugins_count = len(({
        plugin_name for plugin_name in discovered['plugins'] if plugin_name not in hedra_config['plugins']
    }))

    logger['console'].sync.info(f'Found - {new_graphs_count} - new graphs and - {new_plugins_count} - plugins.')


    logger['console'].sync.info('Saving project state to .hedra.json config.')

    hedra_config['project'] = hedra_project_config
    hedra_config['core'] = {
        "bypass_connection_validation": bypass_connection_validation,
        "connection_validation_retries": connection_validation_retries
    }

    with open(hedra_config_filepath, 'w') as hedra_config_file:
            hedra_config = json.dump(hedra_config, hedra_config_file, indent=4)
            
    logger['console'].sync.info('Project fetch complete!\n')


from typing import Any
from hedra.core.graphs.hooks.registry.registry_types import EventHook
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook


class Event:

    def __init__(self, target: Hook, source: EventHook) -> None:
        self.target = target
        self.source = source
        self.target_name = None
        self.target_shortname = None

        if target:
            self.target_name = self.target.name
            self.target_shortname = self.target.shortname

        self.pre = source.pre
        self.as_hook = False

    def __getattribute__(self, name: str) -> Any:
        target = object.__getattribute__(self, 'target')
        source = object.__getattribute__(self, 'source')
        if hasattr(target, name) and name != 'call':
            return getattr(target, name)

        elif target is None and hasattr(source, name) and name != 'call':
            return getattr(source, name)

        return object.__getattribute__(self, name)

    async def call(self, *args, **kwargs):
        if self.source.pre is True and self.target:
            await self.source.call()
            result = await self.target.call(*args, **kwargs)

        elif self.source.pre is False and self.target:
            result = await self.target.call(*args, **kwargs)
            await self.source.call(result)

        else:
            result = await self.source.call()

        return result


import os
from types import SimpleNamespace
from typing import Dict
from pydantic import BaseModel
from pydantic.dataclasses import dataclass

from hedra.reporting.types.common.types import ReporterTypes

try:
    import sqlalchemy
except Exception:
    sqlalchemy = SimpleNamespace(Column=None)


class SQLiteConfig(BaseModel):
    path: str=f'{os.getcwd()}/results.db'
    events_table: str='events'
    metrics_table: str='metrics'
    custom_fields: Dict[str, sqlalchemy.Column]={}
    reporter_type: ReporterTypes=ReporterTypes.SQLite

    class Config:
        arbitrary_types_allowed = True


import asyncio
import functools
import json
import psutil
import uuid
from typing import List
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet
from .json_config import JSONConfig

has_connector = True

class JSON:

    def __init__(self, config: JSONConfig) -> None:
        self.events_filepath = config.events_filepath
        self.metrics_filepath = config.metrics_filepath
        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self._loop = asyncio.get_event_loop()

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()


    async def connect(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping connect')

    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Events to file - {self.events_filepath}')

        event_records = [
            event.record for event in events
        ]

        with open(self.events_filepath, 'w') as events_file:
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    json.dump,
                    event_records, 
                    events_file, 
                    indent=4
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Events to file - {self.events_filepath}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping Shared Metrics')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Metrics to file - {self.metrics_filepath}')

        records = {}
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            
            groups = metrics_set.custom_metrics

            for group_name, group in metrics_set.groups.items():
                groups[group_name] = group.record

            records[metrics_set.name] = {
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                'errors': metrics_set.errors,
                **metrics_set.common_stats,
                'groups': groups
            }

        with open(self.metrics_filepath, 'w') as metrics_file:
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    json.dump,
                    records, 
                    metrics_file, 
                    indent=4
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Metrics to file - {self.metrics_filepath}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping Custom Metrics')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping Error Metrics')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')


from hedra.tools.data_structures.async_list import AsyncList


class JobQueue:

    def __init__(self, batch_size, job_workers) -> None:
        self.batch_size = batch_size
        self.workers = job_workers
        self.jobs_count = job_workers.count
        self.job_sizes = []
        self._job_sizes = AsyncList()
        self.calculated = False

    async def setup_jobs(self):
        job_size = int(self.batch_size / self.jobs_count)
        remainder = 0
        self.job_sizes = []

        if self.batch_size % self.jobs_count:
            remainder = self.batch_size % self.jobs_count

        for _ in range(self.jobs_count):
            self.job_sizes += [job_size]

        self.job_sizes[self.jobs_count - 1] += remainder

        self._job_sizes.data = self.job_sizes
        self.calculated

        return self.job_sizes

    async def map_jobs(self, group_on='address'):
        job_size = int(self.batch_size / self.jobs_count)
        remainder = 0
        self.job_sizes = {}

        if self.batch_size % self.jobs_count:
            remainder = self.batch_size % self.jobs_count

        for idx, worker in enumerate(self.workers):

            if group_on == 'server':
                key = worker.__getattribute__('server')
            
            else:
                key = worker.__getattribute__('address')
            
            self.job_sizes[key] = job_size

            if idx == (self.jobs_count - 1):
                self.job_sizes[key] += remainder

        return self.job_sizes

    async def not_empty(self):
        return await self._job_sizes.size() > 0

    async def get_job(self):
        return await self._job_sizes.pop()


import asyncio
from typing import Generic, TypeVar, List, Union
from hedra.core.graphs.events import Event
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.hooks.hook_types.internal import Internal
from hedra.core.graphs.hooks.registry.registry_types import (
    EventHook,
    ContextHook
)
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.plugins.types.plugin_types import PluginType
from hedra.reporting import (
    Reporter,
    BigQueryConfig,
    BigTableConfig,
    CassandraConfig,
    CloudwatchConfig,
    CosmosDBConfig,
    CSVConfig,
    DatadogConfig,
    DogStatsDConfig,
    GoogleCloudStorageConfig,
    GraphiteConfig,
    HoneycombConfig,
    InfluxDBConfig,
    JSONConfig,
    KafkaConfig,
    MongoDBConfig,
    MySQLConfig,
    NetdataConfig,
    NewRelicConfig,
    PostgresConfig,
    PrometheusConfig,
    RedisConfig,
    S3Config,
    SnowflakeConfig,
    SQLiteConfig,
    StatsDConfig,
    TelegrafConfig,
    TelegrafStatsDConfig,
    TimescaleDBConfig
)
from hedra.core.graphs.stages.base.stage import Stage


T = TypeVar('T')


class Submit(Stage, Generic[T]):
    stage_type=StageTypes.SUBMIT
    config: (
        BigQueryConfig | BigTableConfig | CassandraConfig | CloudwatchConfig | CosmosDBConfig |
        CSVConfig | DatadogConfig | DogStatsDConfig | GoogleCloudStorageConfig | GraphiteConfig | HoneycombConfig |
        InfluxDBConfig | JSONConfig | KafkaConfig | MongoDBConfig | MySQLConfig | NetdataConfig |
        NewRelicConfig | PostgresConfig | PrometheusConfig | RedisConfig | S3Config | SnowflakeConfig |
        SQLiteConfig | StatsDConfig | TelegrafConfig | TelegrafStatsDConfig | TimescaleDBConfig
    ) = None
    submit_events: bool = False
    
    def __init__(self) -> None:
        super().__init__()
        self.summaries = {}
        self.events = []
        self.reporter: Reporter = None
        self.accepted_hook_types = [ HookType.EVENT, HookType.CONTEXT ]

    @Internal()
    async def run(self):

        events: List[Union[EventHook, Event]] = [event for event in self.hooks[HookType.EVENT]]
        pre_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre
        ]
        
        if len(pre_events) > 0:
            pre_event_names = ", ".join([
                event.shortname for event in pre_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing PRE events - {pre_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in pre_events
            ], timeout=self.stage_timeout)


        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Initializing results submission')

        reporter_plugins = self.plugins_by_type.get(PluginType.REPORTER)

        for plugin_name, plugin in reporter_plugins.items():
            Reporter.reporters[plugin_name] = plugin

            if isinstance(self.config, plugin.config):
                self.config.reporter_type = plugin_name
            
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Loaded Reporter plugin - {plugin_name}')
        
        self.reporter = Reporter(self.config)
        self.reporter.graph_name = self.graph_name
        self.reporter.graph_id = self.graph_id
        self.reporter.stage_name = self.name
        self.reporter.stage_id = self.stage_id
        
        reporter_name = self.reporter.reporter_type.name.capitalize()

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Submitting results via - {reporter_name}:{self.reporter.reporter_id} - reporter')
        await self.logger.spinner.append_message(f'Submitting results via - {reporter_name} - reporter')

        await self.reporter.connect()

        session_total = self.summaries.get('session_total', 0)
        if self.submit_events:

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {reporter_name}:{self.reporter.reporter_id} - Submitting - {session_total} - Events')
            await self.reporter.submit_events(self.events)

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {reporter_name}:{self.reporter.reporter_id} - Submitted - {session_total} - Events')

        metrics = []

        stage_summaries = self.summaries.get('stages', {})
        for stage in stage_summaries.values():
            metrics.extend(list(
                stage.get('actions', {}).values()
            ))

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {reporter_name}:{self.reporter.reporter_id} - Submitting Metrics')

        await self.reporter.submit_common(metrics)
        await self.reporter.submit_metrics(metrics)
        await self.reporter.submit_errors(metrics)
        await self.reporter.submit_custom(metrics)
        await self.reporter.close()

        post_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre is False
        ]

        if len(post_events) > 0:
            post_event_names = ", ".join([
                event.shortname for event in post_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing POST events - {post_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in post_events
            ], timeout=self.stage_timeout)

        context_hooks: List[ContextHook] = self.hooks[HookType.CONTEXT]
        context_hooks: List[ContextHook] = self.hooks[HookType.CONTEXT]
        await asyncio.gather(*[
            asyncio.create_task(context_hook.call(self.context)) for context_hook in context_hooks
        ])

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Reporter - {reporter_name}:{self.reporter.reporter_id} - Completed Metrics submission')
        await self.logger.spinner.set_default_message(f'Successfully submitted the results for {session_total} actions via {reporter_name} reporter')
        


import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    BigTableConfig
)


class SubmitBigTableResultsStage(Submit):
    config=BigTableConfig(
        service_account_json_path=os.getenv('GOOGLE_CLOUD_ACCOUNT_JSON_PATH', ''),
        instance_id=os.getenv('GOOGLE_CLOUD_BIGTABLE_INSTANCE_ID', ''),
        events_table='events',
        metrics_table='metrics'
    )


import time
from hedra.tools.helpers import check_event_loop
from statserve.server import Server


class EmbeddedStatserve:

    def __init__(self, config):
        self.config = config
        self.server = Server(
            streams=self.config.distributed_config.get(
                'streams',
                [
                    {
                        'stream_name': 'hedra'
                    }
                ]
            ),
            config=self.config.distributed_config
        )
        self.running = False
        self.init_wait = self.config.distributed_config.get('init_wait', 10)

    def run(self):
        self.running = True
        self.server.start(blocking=False)
        time.sleep(self.init_wait)

    def kill(self):
        self.server.stop()


import asyncio
import os
import grpc

from .worker_registry import WorkerRegistry
from hedra.runners.utils.connect_timeout import connect_or_return_none, connect_with_retry_async
from hedra.runners.worker_services.proto import (
    WorkerHeartbeatRequest,
    LeaderRegisterWorkerRequest,
    LeaderUpdateRequest,
    NewJobRequest,
    JobCompleteRequest
)
from google.protobuf.json_format import (
    MessageToDict
)



class WorkerServicesManager:

    def __init__(self) -> None:
        logger = Logger()
        self.session_logger = logger.generate_logger()
        self.leader_ip = None
        self.leader_port = None
        self.leader_id = None
        self.leader_address = None
        self.worker_ips = None
        self.worker_ports = None
        self.worker_request_timeout = None
        self.worker_addresses = None
        self.registry = None

    async def initialize(self):
        await self.registry.initialize()

    async def registered(self):
        return self.registry.worker_addresses

    async def setup(self, config, leader_id):
        self.leader_id = leader_id

        self.worker_ips = config.distributed_config.get(
            'worker_server_ips',
            os.getenv('WORKER_SERVER_IPS')
        )

        self.worker_ports = config.distributed_config.get(
            'worker_server_ports',
            os.getenv('WORKER_SERVER_PORTS', "6671")
        )

        self.worker_request_timeout =config.distributed_config.get(
            'worker_request_timeout',
            30
        )

        worker_addresses = config.distributed_config.get(
            'worker_server_addresses',
            os.getenv('WORKER_SERVER_ADDRESSES')
        )

        workers_count = config.distributed_config.get(
            'workers',
            os.getenv('WORKERS')
        )

        if workers_count:
            self.worker_addresses = []
        
        else:
            if worker_addresses and len(worker_addresses) > 0:
                worker_addresses = worker_addresses.split(",")
                self.worker_addresses = worker_addresses

            elif self.worker_ips and len(self.worker_ips) > 0:
                self.worker_ips = self.worker_ips.split(",")
                self.worker_ports = [int(port) for port in self.worker_ports.split(",")]

                self.worker_addresses = []

                for worker_ip, worker_port in zip(self.worker_ips, self.worker_ports):
                    self.worker_addresses.append(
                        f'{worker_ip}:{worker_port}'
                    )
            
            else:
                self.worker_addresses = []

        worker_address_strings = ''.join([f'\n- {worker_address}' for worker_address in self.worker_addresses])
        
        self.session_logger.info(f'\nConnecting to workers at addresses: {worker_address_strings}\n')

        self.registry = WorkerRegistry(
            self.worker_addresses,
            workers_count=workers_count
        )

    async def add_new_workers(self, new_workers_queue):
        
        has_new_workers = await new_workers_queue.not_empty()
        if has_new_workers:        
            await self.register_leader(workers=new_workers_queue)

        return new_workers_queue

    async def leader_register_worker(self, workers=None):
        if workers is None:
            workers = self.registry

        responses, _ = await asyncio.wait([
            self._leader_register_worker(
                workers[worker_address]
            ) for worker_address in workers.worker_addresses
        ], timeout=self.worker_request_timeout)

        results = await asyncio.gather(*responses)
        
        for registration_response in results:
            if registration_response:
                await self.registry.register_worker(registration_response.host_address)

    @connect_with_retry_async(wait_buffer=5, timeout_threshold=15)
    async def _leader_register_worker(self, worker):
        request =  LeaderRegisterWorkerRequest(
            host_address=self.leader_address,
            host_id=str(self.leader_id)
        )

        return await worker.client.LeaderRegisterWorker(request)

    async def get_actions_completed(self):
        self.session_logger.debug('Getting completed requests...')

        worker_completed_responses = await asyncio.wait([
            self._get_actions_completed(
                self.registry[worker_address],
                worker_address
            ) for worker_address in self.registry.worker_addresses
        ], timeout=self.worker_request_timeout)

        return await asyncio.gather(*worker_completed_responses)

    @connect_with_retry_async(wait_buffer=5,timeout_threshold=15)
    async def _get_actions_completed(self, worker):
        try:
            leader_update_request = LeaderUpdateRequest(
                host_address=self.leader_address,
                host_id=str(self.leader_id)
            )
            update_response = await worker.client.GetRequestsCompleted(leader_update_request)

            return MessageToDict(update_response)

        except Exception as err:
            return {}

    async def create_new_worker_job(self, job):
        worker_job_responses, _ = await asyncio.wait([
            self._create_new_worker_job(
                self.registry[worker_address],
                job,
                worker_address
            ) for worker_address in self.registry.registered_workers
        ], timeout=self.worker_request_timeout)

        return await asyncio.gather(*worker_job_responses)

    @connect_with_retry_async(wait_buffer=5, timeout_threshold=15)
    async def _create_new_worker_job(self, worker, job, worker_address):
        request = NewJobRequest(
                host_address=self.leader_address,
                host_id=str(self.leader_id),
                job_id=job.job_id,
                job_timeout=job.timeout
            )

        request.job_config.update(job.pipeline.configs[worker_address])
        return await worker.client.CreateNewWorkerJob(request)

    async def get_worker_job_results(self, job):
        worker_job_responses, _ = await asyncio.wait([
            self._get_worker_job_results(
                self.registry[worker_address],
                job
            ) for worker_address in self.registry.registered_workers
        ], timeout=self.worker_request_timeout)

        results = await asyncio.gather(*worker_job_responses)
        parsed_results = [MessageToDict(result) for result in results]

        aggregated_events = []
        
        for parsed_result in parsed_results:
            aggregated_events.extend(parsed_result.get('jobResults').get('aggregated'))

        return aggregated_events

    @connect_with_retry_async(wait_buffer=5, timeout_threshold=15)
    async def _get_worker_job_results(self, worker, job):

        request = JobCompleteRequest(
            host_address=self.leader_address,
            host_id=str(self.leader_id),
            job_id=job.job_id
        )

        return await worker.client.GetJobResults(request)

    async def check_worker_heartbeat(self):

        self.session_logger.debug('Getting worker heartbeats...')

        for worker in self.registry:
            worker_heartbeat_response = await self._check_worker_heartbeat(worker)
            await self._update_registry(worker_heartbeat_response, worker.worker_address)
            
    @connect_or_return_none(wait_buffer=1, timeout_threshold=3)
    async def _check_worker_heartbeat(self, worker):
        request = WorkerHeartbeatRequest(
            host_address=self.leader_address,
            host_id=str(self.leader_id),
            status='OK'
        )

        return await worker.client.CheckWorkerHeartbeat(request)

    async def _update_registry(self, worker_response, worker_address):

        if worker_response and isinstance(worker_response, grpc.aio.AioRpcError) is False:
            
            await self.registry.register_worker(worker_response.host_address)

            id_not_set = await self.registry.worker_id_not_set(worker_response.host_address)
            if id_not_set:
                await self.registry.set_worker_id(worker_response.host_address, worker_response.host_id)

        else:
            await self.registry.remove_worker(worker_address)



import os
from .service_registry import ServiceRegistry
import uuid
from .proto import (
    ServiceRegisterRequest,
    ServiceRegistrationStatusRequest,
    ServiceRegistrationStatusResponse,
    BootstrapServerServicer,

)

from broadkast.grpc import BroadkastServer
from google.protobuf.json_format import (
    MessageToDict
)


class BootstrapService(BootstrapServerServicer):

    def __init__(self, config) -> None:
        super().__init__()

        self.bootstrap_service_id = str(uuid.uuid4())

        self.bootstrap_ip = config.distributed_config.get(
            'bootstrap_ip',
            os.getenv('BOOSTRAP_SERVER_IP')
        )

        self.bootstrap_port = config.distributed_config.get(
            'bootstrap_port',
            os.getenv('BOOTSTRAP_SERVER_PORT', "8711")
        )

        self.bootstrap_service_address = config.distributed_config.get(
            'bootstrap_address',
            os.getenv(
                'BOOTSTRAP_SERVER_ADDRESS', 
                f'{self.bootstrap_ip}:{self.bootstrap_port}'
            )
        )

        expected_leaders = int(config.distributed_config.get(
            'leaders', 
            os.getenv('LEADERS', 0)
        ))

        self.registry = ServiceRegistry(expected_leaders)
        self.broadkast_server = BroadkastServer(config=config.distributed_config)

        self.running = True

    async def RegisterService(self, request, context):
        register_request = MessageToDict(request)
        
        registration_response = await self.registry.register_service(register_request)

        service_registration_status_response = ServiceRegistrationStatusResponse(
            bootstrap_service_address=self.bootstrap_service_address,
            bootstrap_service_id=self.bootstrap_service_id,
            registered_count=registration_response.get('registered_count'),
            ready=registration_response.get('ready')
        )

        service_registration_status_response.registered_addresses.extend(
            registration_response.get('registered_addresses')
        )

        service_registration_status_response.registered_ids.extend(
            registration_response.get('registered_ids')
        )

        service_registration_status_response.registered_server_addresses.extend(
            registration_response.get('registered_server_addresses')
        )

        return service_registration_status_response

    async def GetServiceRegistrationStatus(self, request, context):

        registration_status_request = MessageToDict(request)
        
        service_registration_status = await self.registry.get_service_registration_status(
            registration_status_request
        )

        service_registration_status_response = ServiceRegistrationStatusResponse(
            bootstrap_service_address=self.bootstrap_service_address,
            bootstrap_service_id=self.bootstrap_service_id,
            registered_count=service_registration_status.get('registered_count'),
            ready=service_registration_status.get('ready')
        )

        service_registration_status_response.registered_addresses.extend(
            service_registration_status.get('registered_addresses')
        )

        service_registration_status_response.registered_ids.extend(
            service_registration_status.get('registered_ids')
        )

        service_registration_status_response.registered_server_addresses.extend(
            service_registration_status.get('registered_server_addresses')
        )

        return service_registration_status_response



import uuid
from typing import Any, Dict, Union


class MetricsGroup:

    def __init__(
        self,
        name: str,
        source: str,
        stage: str, 
        group_name: str,
        data: Dict[str, Union[int, float, str, bool]],
        common: Dict[str, int]
    ) -> None:

        self.metrics_group_id = str(uuid.uuid4())

        self.name = name
        self.stage = stage
        self.group_name = group_name
        self.data = []

        self.fields = [
            'name',
            'stage'
        ]
        self.values = [
            name,
            stage
        ]

        self._additional_fields = ["quantiles", "errors", "custom_metrics"]
        self.source = source
        self._raw_data = data
        self._common = common

        flattened_data = dict({
            field: value for field, value in data.items() if field not in self._additional_fields
        })

    
        for metric_name, metric_value in flattened_data.items():
            self.data.append((
                metric_name,
                metric_value
            ))

            self.fields.append(metric_name)
            self.values.append(metric_value)

        for quantile, quantile_value in data.get("quantiles", {}).items():
            self.data.append((
                quantile,
                quantile_value
            ))

            self.fields.append(quantile)
            self.values.append(quantile_value)
            

        self.unique = list(self.data)
        self.unique_fields = list(self.fields)
        self.unique_values = list(self.values)

        self.custom_fields: Dict[str, Dict[str, Union[int, float, Any]]] = data.get('custom_metrics', {})

    @property
    def custom(self):
        return {
            custom_metric_name: custom_metric.get(
                'result'
            ) for custom_metric_name, custom_metric in self.custom_fields.items()
        }

    @property
    def custom_schemas(self):
        return {
            custom_metric_name: custom_metric.get(
                'field_type'
            ) for custom_metric_name, custom_metric in self.custom_fields.items()
        }

    @property
    def custom_field_names(self):
        return list(self.custom_fields.keys())

    @property
    def record(self):

        record_data = {
            f'{field}': value for field, value in self.data
        }

        custom_field_data = {
            f'{field}': value for field, value in self.custom.items()
        }

        return {
            'name': self.name,
            'stage': self.stage,
            **record_data,
            **custom_field_data
        }

    @property
    def raw_data(self):
        return self._raw_data

    @property
    def stats(self):
        stats_metrics = {}
        for metric_name, metric_value in self.data:
            if isinstance(metric_value, (int, float)):
                stats_metrics[metric_name] = metric_value

        return stats_metrics

    @property
    def quantiles(self):
        return self._raw_data.get('quantiles', {})


import dill
import time
import statistics
import asyncio
from collections import defaultdict
from typing import Union, List, Dict, Union
from hedra.plugins.types.plugin_types import PluginType
from hedra.reporting.events import EventsGroup
from hedra.reporting.metric import MetricsSet
from hedra.core.graphs.events import Event
from hedra.core.graphs.hooks.hook_types.internal import Internal
from hedra.core.graphs.hooks.registry.registry_types import (
    EventHook, 
    ContextHook,
    MetricHook,
)
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.reporting.events import results_types
from hedra.reporting.events.types import (
    HTTPEvent, 
    HTTP2Event, 
    GraphQLEvent, 
    GRPCEvent, 
    WebsocketEvent, 
    PlaywrightEvent,
    UDPEvent
)
from hedra.core.graphs.stages.base.stage import Stage
from .parallel import process_results_batch


Events = Union[HTTPEvent, HTTP2Event, GraphQLEvent, GRPCEvent, WebsocketEvent, PlaywrightEvent, UDPEvent]


class Analyze(Stage):
    stage_type=StageTypes.ANALYZE
    is_parallel = False
    handler = None

    def __init__(self) -> None:
        super().__init__()
        self.raw_results = {}

        self.accepted_hook_types = [ HookType.METRIC, HookType.EVENT, HookType.CONTEXT ]
        self.requires_shutdown = True
        self.allow_parallel = True
        self.analysis_execution_time = 0

    @Internal()
    async def run(self):

        events: List[Union[EventHook, Event]] = [event for event in self.hooks[HookType.EVENT]]
        pre_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre
        ]
        
        if len(pre_events) > 0:
            pre_event_names = ", ".join([
                event.shortname for event in pre_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing PRE events - {pre_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in pre_events
            ], timeout=self.stage_timeout)

        await self.logger.filesystem.aio.create_logfile('hedra.reporting.log')
        self.logger.filesystem.create_filelogger('hedra.reporting.log')
        
        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Starting results analysis')

        analysis_execution_time_start = time.monotonic()

        engine_plugins = self.plugins_by_type.get(PluginType.ENGINE)
        for plugin_name, plugin in engine_plugins.items():
            results_types[plugin_name] = plugin.event

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Generated custom Event - {plugin.event.type} - for Reporter plugin - {plugin_name}')
        
        all_results = list(self.raw_results.items())
        summaries = {
            'session_total': 0,
            'stages': {}
        }

        metric_hook_names: List[str] = [hook.name for hook in self.hooks[HookType.METRIC]]

        batches = self.executor.partion_stage_batches(all_results)
        total_group_results = 0

        elapsed_times = []
        for stage_name, _, _ in batches:
            stage_results = self.raw_results.get(stage_name)
            total_group_results += stage_results.get('total_results', 0)
            elapsed_times.append(
                stage_results.get('total_elapsed', 0)
            )

        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Paritioned {len(batches)} batches of results')

        median_execution_time = round(statistics.median(elapsed_times))
        await self.logger.spinner.append_message(f'Calculating stats for - {total_group_results} - actions executed over a median stage execution time of {median_execution_time} seconds')

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Calculating stats for - {total_group_results} - actions over a median stage execution time of {median_execution_time} seconds')

        custom_metric_hooks: List[MetricHook] = []
        for metric_hook_name in metric_hook_names:
            custom_metric_hook = registrar.all.get(metric_hook_name)
            custom_metric_hooks.append(custom_metric_hook)

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Loaded custom Metric hook - {metric_hook_name}')

        stage_configs = []
        stage_total_times = {}

        for stage_name, _, assigned_workers_count in batches:
            
            stage_batches = []

            stage_results = self.raw_results.get(stage_name)
            results = stage_results.get('results')
            stage_total_time = stage_results.get('total_elapsed')
            
            stage_total_times[stage_name] = stage_total_time

            results_count = len(results)
            
            batch_size = int(results_count/assigned_workers_count)
            
            for worker_idx in range(assigned_workers_count):

                batch_marker = worker_idx * batch_size

                stage_batches.append(
                    results[batch_marker:batch_marker + batch_size]
                )

            if results_count%assigned_workers_count > 0:
                stage_batches[assigned_workers_count-1].extend(
                    results[assigned_workers_count * batch_size:]
                )

            batch_configs = []
            for batch in stage_batches:
                batch_configs.append({
                    'graph_name': self.graph_name,
                    'graph_id': self.graph_id,
                    'source_stage_name': self.name,
                    'source_stage_id': self.stage_id,
                    'analyze_stage_name': stage_name,
                    'analyze_stage_metric_hooks': list(metric_hook_names),
                    'analyze_stage_batched_results': batch
                })

            stage_configs.append((
                stage_name,
                assigned_workers_count,
                batch_configs
            ))

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Assigned {assigned_workers_count} to process results from stage - {stage_name}')

        stages_count = len(stage_configs)
        await self.logger.spinner.append_message(
            f'Calculating results for - {stages_count} - stages'
        )

        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Processing results or - {stages_count} - stages')


        stage_batches = await self.executor.execute_batches(
            stage_configs,
            process_results_batch
        )

        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Completed parital results aggregation for - {stages_count} - stages')

        processed_results = []

        self.logger.spinner.set_message_at(2, f'Converting aggregate results to metrics for - {stages_count} - stages.')

        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Starting stage results aggregation for {stages_count} stages')
        
        for stage_name, stage_results in stage_batches:
        
            stage_total = 0
            stage_total_time = stage_total_times.get(stage_name)

            batch_results: List[Dict[str, Union[dict, EventsGroup]]] = [
                dill.loads(group) for group in stage_results
            ]

            events =  defaultdict(EventsGroup)

            for events_groups in batch_results:
                for event_group_name, events_group in events_groups.items():
                    events[event_group_name].merge(events_group)

            grouped_stats = {}
            
            for event_group_name, events_group in events.items():  

                custom_metrics = defaultdict(dict)
                for custom_metric in custom_metric_hooks:
                    custom_metrics[custom_metric.group][custom_metric.shortname] = await custom_metric.call(
                        self.raw_results.get(
                            stage_name
                        ).get('results')
                    )

                events_group.calculate_quantiles()

                metric_data = {
                    'total': events_group.total,
                    'succeeded': events_group.succeeded,
                    'failed': events_group.failed,
                    'actions_per_second': events_group.total/stage_total_time,
                    'errors': list([
                        {
                            'message': error_message,
                            'count': error_count
                        } for error_message, error_count in events_group.errors.items()
                    ]),
                    'groups': events_group.groups,
                    'custom': custom_metrics
                }

                metric = MetricsSet(
                    event_group_name,
                    events_group.source,
                    stage_name,
                    metric_data,
                    events_group.tags
                )

                grouped_stats[event_group_name] = metric

                stage_total += events_group.total

                await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Convererted stats for stage - {stage_name} to metrics set')
        
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Calculated results for - {events_group.total} - actions from stage - {stage_name}')

            processed_results.append({
                'stage_metrics': {
                    stage_name: {
                        'total': stage_total,
                        'actions_per_second': stage_total/stage_total_time,
                        'actions': grouped_stats
                    }
                },
                'stage_total': stage_total
            })

        for result in processed_results:
            summaries['stages'].update(result.get('stage_metrics'))
            summaries['session_total'] += result.get('stage_total')

        self.analysis_execution_time = round(
            time.monotonic() - analysis_execution_time_start
        )

        post_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre is False
        ]

        if len(post_events) > 0:
            post_event_names = ", ".join([
                event.shortname for event in post_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing POST events - {post_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in post_events
            ], timeout=self.stage_timeout)

        context_hooks: List[ContextHook] = self.hooks[HookType.CONTEXT]
        await asyncio.gather(*[
            asyncio.create_task(context_hook.call(self.context)) for context_hook in context_hooks
        ])

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Completed results analysis for - {stages_count} - stages in - {self.analysis_execution_time} seconds')
        await self.logger.spinner.set_default_message(f'Completed results analysis for {total_group_results} actions and {stages_count} stages over {self.analysis_execution_time} seconds')

        return summaries



import uuid
from collections import defaultdict
from typing import Any, List
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet

try:
    from influxdb_client import Point
    from influxdb_client.client.influxdb_client_async import InfluxDBClientAsync
    from .influxdb_config import InfluxDBConfig
    has_connector = True

except Exception:
    Point = None
    InfluxDBClientAsync = None
    InfluxDBConfig = None
    has_connector = False


class InfluxDB:

    def __init__(self, config: InfluxDBConfig) -> None:
        self.host = config.host
        self.token = config.token
        self.protocol = 'https' if config.secure else 'http'
        self.organization = config.organization
        self.connect_timeout = config.connect_timeout
        self.events_bucket_name = config.events_bucket
        self.metrics_bucket_name = config.metrics_bucket
        self.shared_metrics_bucket_name = 'stage_metrics'
        self.errors_bucket_name = 'stage_errors'
        self.events_database = None
        self.metrics_database = None

        self.client = None
        self.write_api = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to InfluxDB at - {self.protocol}://{self.host} - for Organization - {self.organization}')

        self.client = InfluxDBClientAsync(
            f'{self.protocol}://{self.host}',
            token=self.token,
            org=self.organization,
            timeout=self.connect_timeout
        )

        self.write_api = self.client.write_api()

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to InfluxDB at - {self.protocol}://{self.host} - for Organization - {self.organization}')

    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Bucket - {self.events_bucket_name}')

        points = []
        for event in events:
            point = Point(event.name)


            for tag in event.tags:
                point.tag(tag.name, tag.value)

            point.field(f'{event.name}_time', event.time)

            if event.success:
                point.field(f'{event.name}_success', 1)
            
            else:
                point.field(f'{event.name}_failed', 1)

            points.append(point)

        await self.write_api.write(
            bucket=self.events_bucket_name,
            record=points
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Bucket - {self.events_bucket_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Bucket - {self.shared_metrics_bucket_name}')

        points = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            point = Point(metrics_set.name)

            for tag in metrics_set.tags:
                point.tag(tag.name, tag.value)

            metric_record = {
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                **metrics_set.common_stats
            }

            for field, value in metric_record.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metric - {metrics_set.name}:common:{field}')
                point.field(field, value)

            points.append(point)

        await self.write_api.write(
            bucket=self.shared_metrics_bucket_name,
            record=points
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Bucket - {self.shared_metrics_bucket_name}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Bucket - {self.metrics_bucket_name}')

        points = []
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            
            for group_name, group in metrics_set.groups.items():
                point = Point(metrics_set.name)

                for tag in metrics_set.tags:
                    point.tag(tag.name, tag.value)

                metric_record = {
                    **group.stats, 
                    **group.custom,
                    'group': group_name
                }

                for field, value in metric_record.items():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metric - {metrics_set.name}:{group_name}:{field}')
                    point.field(field, value)

                points.append(point)

        await self.write_api.write(
            bucket=self.metrics_bucket_name,
            record=points
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Bucket - {self.metrics_bucket_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        points = defaultdict(list)
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for custom_group_name, group in metrics_set.custom_metrics.items():

                point = Point(f'{metrics_set.name}_{custom_group_name}')

                for tag in metrics_set.tags:
                    point.tag(tag.name, tag.value)
                
                metric_record = {
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': custom_group_name,
                    **group
                }

                for field, value in metric_record.items():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metric - {metrics_set.name}:{group_name}:{field}')
                    point.field(field, value)

                points[custom_group_name].append(point)

        for group_name, points in points.items():
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to Bucket - {group_name}_metrics')
            await self.write_api.write(
                bucket=f'{group_name}_metrics',
                record=points
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Bucket - {group_name}_metrics')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Errors Metrics to Bucket - {self.errors_bucket_name}')

        points = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Errors Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                point = Point(f'{metrics_set.name}_errors')
                point.field(
                    error.get('message'),
                    error.get('count')
                )

            points.append(point)

        await self.write_api.write(
            bucket=self.errors_bucket_name,
            record=points
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Errors Metrics to Bucket - {self.errors_bucket_name}')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connectiion to InfluxDB at - {self.protocol}://{self.host} - for Organization - {self.organization}')
        
        await self.client.close()

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Session Closed - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connectiion to InfluxDB at - {self.protocol}://{self.host} - for Organization - {self.organization}')



import time
import logging
import functools
import os
import asyncio
import grpc


def connect_with_retry(wait_buffer=5,timeout_threshold=int(os.getenv('GPRC_CONNECT_TIMEOUT', 30))): 
    def retry(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            start = time.time()
            elapsed = 0
            result = None
            while elapsed < timeout_threshold:
                try:
                    result = func(*args, **kwargs)
                    break
                except Exception as error:
                    result = None
                    logging.error('Error: Connection attempt failed. Retrying in {wait_buffer}'.format(wait_buffer=wait_buffer))
                    logging.debug(str(error))
                if result:
                    return result

                time.sleep(wait_buffer)
                current = time.time()
                elapsed = current - start

            return result
        return wrapper
    return retry


def connect_with_retry_async(wait_buffer=5,timeout_threshold=int(os.getenv('GPRC_CONNECT_TIMEOUT', 30))): 
    def retry(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            start = time.time()
            elapsed = 0
            result = None
            while elapsed < timeout_threshold:
                try:
                    result = await func(*args, **kwargs)
                    break
                except Exception as error:
                    result = None
                    logging.error('Error: Connection attempt failed. Retrying in {wait_buffer}'.format(wait_buffer=wait_buffer))
                    logging.debug(str(error))
                if result:
                    return result

                await asyncio.sleep(wait_buffer)
                current = time.time()
                elapsed = current - start

            return result
        return wrapper
    return retry



def connect_or_return_none(wait_buffer=5,timeout_threshold=int(os.getenv('GPRC_CONNECT_TIMEOUT', 30))): 
    def retry(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            start = time.time()
            elapsed = 0
            result = None
            while elapsed < timeout_threshold:
                try:
                    result = await func(*args, **kwargs)

                    if isinstance(result, grpc.aio.AioRpcError):
                        result = None

                    break
                except Exception as error:
                    result = None
                if result:
                    return result

                await asyncio.sleep(wait_buffer)
                current = time.time()
                elapsed = current - start

            return result
        return wrapper
    return retry


from collections import defaultdict
from types import FunctionType
from typing import Any, Dict
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from .registry_types import (
    ActionHook,
    AfterHook,
    BeforeHook,
    ChannelHook,
    CheckHook,
    ContextHook,
    EventHook,
    MetricHook,
    RestoreHook,
    SaveHook,
    SetupHook,
    TaskHook,
    TeardownHook,
    ValidateHook
)




class Registrar:
    all: Dict[str, Hook] = {}
    reserved: Dict[str, Dict[str, Hook]] = defaultdict(dict)
    module_paths: Dict[str, str] = {}

    def __init__(self, hook_type) -> None:
        self.hook_type = hook_type
        self.hook_types = {
            HookType.ACTION: lambda *args, **kwargs: ActionHook(*args, **kwargs),
            HookType.AFTER: lambda *args, **kwargs:  AfterHook(*args, **kwargs),
            HookType.BEFORE: lambda *args, **kwargs:  BeforeHook(*args, **kwargs),
            HookType.CHANNEL: lambda *args, **kwargs:  ChannelHook(*args, **kwargs),
            HookType.CHECK: lambda *args, **kwargs:  CheckHook(*args, **kwargs),
            HookType.CONTEXT: lambda *args, **kwargs:  ContextHook(*args, **kwargs),
            HookType.EVENT: lambda *args, **kwargs:  EventHook(*args, **kwargs),
            HookType.METRIC: lambda *args, **kwargs:  MetricHook(*args, **kwargs),
            HookType.RESTORE: lambda *args, **kwargs:  RestoreHook(*args, **kwargs),
            HookType.SAVE: lambda *args, **kwargs:  SaveHook(*args, **kwargs),
            HookType.SETUP: lambda *args, **kwargs:  SetupHook(*args, **kwargs),
            HookType.TASK: lambda *args, **kwargs:  TaskHook(*args, **kwargs),
            HookType.TEARDOWN: lambda *args, **kwargs:  TeardownHook(*args, **kwargs),
            HookType.VALIDATE: lambda *args, **kwargs:  ValidateHook(*args, **kwargs)
        }

    def __call__(self, hook: FunctionType) -> Any:
        self.module_paths[hook.__name__] = hook.__module__
        return self.add_hook(self.hook_type)

    def add_hook(self, hook_type: str):

        def wrap_hook(*args, **kwargs):

            def wrapped_method(func):

                hook_name = func.__qualname__
                hook_shortname = func.__name__

                hook = self.hook_types[hook_type]

                hook_args = args
                args_count = len(args)
                
                if args_count < 1:
                    hook_args = []

                self.all[hook_name] = hook(
                    hook_name,
                    hook_shortname,
                    func,
                    *hook_args,
                    **kwargs
                )

                return func
            
            return wrapped_method

        return wrap_hook
        

def makeRegistrar():
    return Registrar


registrar = makeRegistrar()


import asyncio
from typing import Coroutine, Dict, List, Tuple
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.task import MercuryTaskRunner, Task, TaskResult
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common import Timeouts
from hedra.logging import HedraLogger
from .base_client import BaseClient


class TaskClient(BaseClient[MercuryTaskRunner, Task, TaskResult]):

    def __init__(self, config: Config) -> None:
        super().__init__()

        self.session = MercuryTaskRunner(
            concurrency=config.batch_size,
            timeouts=Timeouts(
                total_timeout=config.request_timeout
            )
        )
        self.request_type = RequestTypes.TASK
        self.client_type = self.request_type.capitalize()
        self.next_name = None

        self.logger = HedraLogger()
        self.logger.initialize()

    def call(
        self, 
        task: Coroutine, 
        env: str = None,
        user: str = None,
        tags: List[Dict[str, str]] = []
    ) -> Tuple[Task, MercuryTaskRunner]:
        task_action = Task(
            self.next_name,
            task,
            source=env,
            user=user,
            tags=tags
        )

        return task_action, self.session


import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    BigQueryConfig
)


class SubmitBigQueryResultsStage(Submit):
    config=BigQueryConfig(
        service_account_json_path=os.getenv('GOOGLE_CLOUD_ACCOUNT_JSON_PATH', ''),
        project_name='test',
        dataset_name='results',
        events_table='events',
        metrics_table='metrics'
    )


from .base_event import BaseEvent


class StreamEnded(BaseEvent):
    event_type='STREAM_ENDED'
    __slots__ = ('stream_id')
    """
    The StreamEnded event is fired whenever a stream is ended by a remote
    party. The stream may not be fully closed if it has not been closed
    locally, but no further data or headers should be expected on that stream.
    """
    def __init__(self):
        #: The Stream ID of the stream that was closed.
        self.stream_id = None

    def __repr__(self):
        return "<StreamEnded stream_id:%s>" % self.stream_id



import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    NewRelicConfig
)


class SubmitNewrelicResultsStage(Submit):
    config=NewRelicConfig(
        config_path=os.getenv('NEWRELIC_CONFIG_PATH', ''),
        environment=os.getenv('NEWRELIC_ENVIRONMENT', ''),
        newrelic_application_name='results'
    )


import uuid
import json
from typing import List, Dict, Any
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet

try:

    from aiokafka import AIOKafkaProducer
    from .kafka_config import KafkaConfig
    has_connector = True

except Exception:
    AIOKafkaProducer = None
    KafkaConfig = None
    has_connector = False


class Kafka:

    def __init__(self, config: KafkaConfig) -> None:
        self.host = config.host
        self.client_id = config.client_id

        self.events_topic = config.events_topic
        self.metrics_topic = config.metrics_topic
        self.custom_metrics_topics = {}
        self.shared_metrics_topic = 'stage_metrics'
        self.errors_topic = 'stage_errors'

        self.events_partition = config.events_partition
        self.metrics_partition = config.metrics_partition
        self.shared_metrics_partition = 'stage_metrics'
        self.errors_partition = 'stage_errors'

        self.compression_type = config.compression_type
        self.timeout = config.timeout
        self.enable_idempotence = config.idempotent or True
        self.options: Dict[str, Any] = config.options or {}
        self._producer = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to Kafka at - {self.host}')

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Using Kafka Options - Compression Type: {self.compression_type}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Using Kafka Options - Connection Timeout: {self.timeout}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Using Kafka Options - Idempotent: {self.enable_idempotence}')

        for option_name, option in self.options.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Using Kafka Options - {option_name.capitalize()}: {option}')


        self._producer = AIOKafkaProducer(
            bootstrap_servers=self.host,
            client_id=self.client_id,
            compression_type=self.compression_type,
            request_timeout_ms=self.timeout,
            enable_idempotence=self.enable_idempotence,
            **self.options
        )

        await self._producer.start()

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to Kafka at - {self.host}')

    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Topic - {self.events_topic} - Partition - {self.events_partition}')

        batch = self._producer.create_batch()
        for event in events:

            batch.append(
                value=json.dumps(
                    event.record
                ).encode('utf-8'),
                timestamp=None, 
                key=bytes(event.name, 'utf')
            )

        await self._producer.send_batch(
            batch,
            self.events_topic,
            partition=self.events_partition
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Topic - {self.events_topic} - Partition - {self.events_partition}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Topic - {self.shared_metrics_topic} - Partition - {self.shared_metrics_partition}')

        batch = self._producer.create_batch()
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            batch.append(
                value=json.dumps({
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'common',
                    **metrics_set.common_stats
                }).encode('utf-8'),
                timestamp=None, 
                key=bytes(metrics_set.name, 'utf')
            )

        await self._producer.send_batch(
            batch,
            self.shared_metrics_topic,
            partition=self.shared_metrics_partition
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Topic - {self.shared_metrics_topic} - Partition - {self.shared_metrics_partition}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Topic - {self.metrics_topic} - Partition - {self.metrics_partition}')
        
        batch = self._producer.create_batch()
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
                batch.append(
                    value=json.dumps(
                        {
                            **group.record,
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': group_name
                        }
                    ).encode('utf-8'),
                    timestamp=None, 
                    key=bytes(f'{metrics_set.name}_{group_name}', 'utf')
                )

        await self._producer.send_batch(
            batch,
            self.metrics_topic,
            partition=self.metrics_partition
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Topic - {self.metrics_topic} - Partition - {self.metrics_partition}')
    
    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Customm Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for custom_group_name, group in metrics_set.custom_metrics.items():
                custom_topic_name = f'{custom_group_name}_metrics'

                if self.custom_metrics_topics.get(custom_topic_name) is None:
                    self.custom_metrics_topics[custom_topic_name] = self._producer.create_batch()
                
                self.custom_metrics_topics[custom_topic_name].append(
                    value=json.dumps({
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'group': custom_group_name,
                        **group
                    }).encode('utf-8'),
                    timestamp=None,
                    key=bytes(f'{metrics_set.name}_{custom_group_name}', 'utf')
                )

        for topic_name, batch in self.custom_metrics_topics.items():
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Customm Metrics to Topic - {topic_name} - Partition - {topic_name}')
            await self._producer.send_batch(
                batch,
                topic_name,
                partition=topic_name
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Customm Metrics to Topic - {topic_name} - Partition - {topic_name}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Topic - {self.metrics_topic} - Partition - {self.metrics_partition}')

        batch = self._producer.create_batch()
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                batch.append(
                    value=json.dumps(
                        {
                            'metric_name': metrics_set.name,
                            'metric_stage': metrics_set.stage,
                            'error_message': error.get('message'),
                            'error_count': error.get('count')
                        }
                    ).encode('utf-8'),
                    timestamp=None, 
                    key=bytes(metrics_set.name, 'utf')
                )
        
        await self._producer.send_batch(
            batch,
            self.errors_topic,
            partition=self.errors_partition
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Topic - {self.metrics_topic} - Partition - {self.metrics_partition}')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connection to Kafka at - {self.host}')

        await self._producer.stop()

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connection to Kafka at - {self.host}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Session Closed - {self.session_uuid}')


import socket


class SocketTypes:
    DEFAULT=socket.AF_INET
    HTTP2=socket.AF_INET
    UDP=socket.AF_INET
    NONE=None

class SocketProtocols:
    DEFAULT=socket.SOCK_STREAM
    HTTP2=socket.SOCK_STREAM
    UDP=socket.SOCK_DGRAM
    NONE=None


class RequestTypes:
    HTTP='HTTP'
    HTTP2='HTTP2'
    WEBSOCKET='WEBSOCKET'
    GRAPHQL='GRAPHQL'
    GRAPHQL_HTTP2="GRAPHQL_HTTP2"
    GRPC='GRPC'
    PLAYWRIGHT='PLAYWRIGHT'
    UDP='UDP'
    TASK='TASK'
    CUSTOM='CUSTOM'


class ProtocolMap:

    def __init__(self) -> None:
        self.address_families = {
            RequestTypes.HTTP: SocketTypes.DEFAULT,
            RequestTypes.HTTP2: SocketTypes.HTTP2,
            RequestTypes.WEBSOCKET: SocketTypes.DEFAULT,
            RequestTypes.GRAPHQL: SocketTypes.DEFAULT,
            RequestTypes.GRAPHQL_HTTP2: SocketTypes.HTTP2,
            RequestTypes.GRPC: SocketTypes.HTTP2,
            RequestTypes.UDP: SocketTypes.UDP,
            RequestTypes.PLAYWRIGHT: SocketTypes.NONE
        }

        self.protocols = {
            RequestTypes.HTTP: SocketProtocols.DEFAULT,
            RequestTypes.HTTP2: SocketProtocols.HTTP2,
            RequestTypes.WEBSOCKET: SocketProtocols.DEFAULT,
            RequestTypes.GRAPHQL: SocketProtocols.DEFAULT,
            RequestTypes.GRAPHQL_HTTP2: SocketTypes.HTTP2,
            RequestTypes.GRPC: SocketProtocols.HTTP2,
            RequestTypes.UDP: SocketProtocols.UDP,
            RequestTypes.PLAYWRIGHT: SocketProtocols.NONE
        }

    def __getitem__(self, key: RequestTypes) -> SocketTypes:
        return self.address_families.get(key), self.protocols.get(key)


from .local import LocalWorker
from .parallel_local_worker import ParallelLocalWorker
from .embedded_statserve import EmbeddedStatserve
from .leader_job_server import LeaderJobServer
from .worker_job_server import DistributedWorkerServer
from .bootstrap_server import BootstrapServer


import functools
from typing import Coroutine, Dict, List, Union
from .hook_type import HookType
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.core.graphs.hooks.registry.registrar import registrar


@registrar(HookType.ACTION)
def action(
    weight: int=1, 
    order: int=1, 
    metadata: Dict[str, Union[str, int]]={}, 
    checks: List[Coroutine]=[], 
    notify: List[str]=[], 
    listen: List[str]=[]
):

    def wrapper(func) -> Hook:

        @functools.wraps(func)
        def decorator(*args, **kwargs):

            return func(*args, **kwargs)
                
        return decorator

    return wrapper


from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    NetdataConfig
)


class SubmitNetdataResultsStage(Submit):
    config=NetdataConfig(
        host='localhost',
        port=8125
    )


import functools
from .hook_type import HookType
from hedra.core.graphs.hooks.registry.registrar import registrar


@registrar(HookType.TEARDOWN)
def teardown(metadata={}):
    '''
    Teardown Hook

    The @teardown(<name>, ...) decorator is an optional means of specifying that
    a method of a class inherting ActionSet should executor after testing as a means
    of cleaning up/closing any session or state created during tests. You may specify as 
    many teardown hooks as needed, however all methods wrapped in the decorator must be 
    asynchronous and return valid results.

    You may pass the following to a teardown hook:

    - name (required:string - positional)
    - group (optional:string - keyword)
    - metadata (optional:dict - keyword - may include env, user, type, tags [list], and url)
    
    '''
    def wrapper(func):

        @functools.wraps(func)
        async def decorator(*args, **kwargs):
            return await func(*args, **kwargs)

        return decorator
    return wrapper


from .base_event import BaseEvent


class RequestReceived(BaseEvent):
    event_type='REQUEST_RECEIVED'
    __slots__ = (
        'stream_id',
        'headers',
        'stream_ended',
        'priority_updated'
    )
    
    """
    The RequestReceived event is fired whenever request headers are received.
    This event carries the HTTP headers for the given request and the stream ID
    of the new stream.

    .. versionchanged:: 2.3.0
       Changed the type of ``headers`` to :class:`HeaderTuple
       <hpack:hpack.HeaderTuple>`. This has no effect on current users.

    .. versionchanged:: 2.4.0
       Added ``stream_ended`` and ``priority_updated`` properties.
    """
    def __init__(self):
        #: The Stream ID for the stream this request was made on.
        self.stream_id = None

        #: The request headers.
        self.headers = None

        #: If this request also ended the stream, the associated
        #: :class:`StreamEnded <h2.events.StreamEnded>` event will be available
        #: here.
        #:
        #: .. versionadded:: 2.4.0
        self.stream_ended = None

        #: If this request also had associated priority information, the
        #: associated :class:`PriorityUpdated <h2.events.PriorityUpdated>`
        #: event will be available here.
        #:
        #: .. versionadded:: 2.4.0
        self.priority_updated = None

    def __repr__(self):
        return "<RequestReceived stream_id:%s, headers:%s>" % (
            self.stream_id, self.headers
        )


import asyncio
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.hooks.hook_types.internal import Internal
from hedra.core.graphs.stages.base.stage import Stage

class Idle(Stage):
    stage_type=StageTypes.IDLE

    def __init__(self) -> None:
        super().__init__()
        self.name = self.__class__.__name__

    @Internal()
    async def run(self):
        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Starting graph execution')
        await asyncio.sleep(1)
        



import asyncio
from typing import Any
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.checkpoint import Checkpoint
from hedra.core.graphs.stages.types.stage_states import StageStates
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.transitions.exceptions import (
    StageExecutionError,
    StageTimeoutError
)
from hedra.logging import HedraLogger


async def teardown_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    teardown_hooks = []
    execute_stages = current_stage.context.stages.get(StageTypes.EXECUTE).values()
    paths = current_stage.context.paths

    valid_states = [
        StageStates.EXECUTED
    ]
    
    if current_stage.state == StageStates.INITIALIZED:

        await logger.spinner.system.debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')
        await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')

        current_stage.state = StageStates.TEARDOWN_INITIALIZED

        for stage in execute_stages:
            in_path = current_stage.name in paths.get(stage.name)

            if stage.state in valid_states and in_path:
                stage.state = StageStates.TEARDOWN_INITIALIZED

                stage_teardown_hooks = stage.hooks[HookType.TEARDOWN]
                if stage_teardown_hooks:
                    teardown_hooks.extend(stage_teardown_hooks)

        current_stage.hooks[HookType.ACTION] = teardown_hooks

        if current_stage.timeout:
            await asyncio.wait_for(current_stage.run(), timeout=current_stage.timeout)

        else:
            await current_stage.run()

        for stage in execute_stages:
            in_path = current_stage.name in paths.get(stage.name)

            if stage.state == StageStates.TEARDOWN_INITIALIZED and in_path:
                stage.state = StageStates.TEARDOWN_COMPLETE

        current_stage.state = StageStates.TEARDOWN_COMPLETE
        
        await logger.spinner.system.debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')
        await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')

    else:
        await logger.spinner.system.debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
        await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')

    next_stage.context = current_stage.context


async def teardown_to_analyze_transition(current_stage: Stage, next_stage: Stage):
    
    try:

        await teardown_transition(current_stage, next_stage)
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    return None, StageTypes.ANALYZE


async def teardown_to_checkpoint_transition(current_stage: Stage, next_stage: Checkpoint):

    try:

        await teardown_transition(current_stage, next_stage)

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR
        
    return None, StageTypes.CHECKPOINT


async def teardown_to_wait_transition(current_stage: Stage, next_stage: Stage):
    
    try:

        await teardown_transition(current_stage, next_stage)
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    return None, StageTypes.WAIT


import asyncio
import traceback
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.types.stage_states import StageStates
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.transitions.exceptions import (
    StageExecutionError,
    StageTimeoutError
)
from hedra.logging import HedraLogger


async def analyze_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    if current_stage.state == StageStates.INITIALIZED:

        await logger.spinner.system.debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')
        await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')
        
        current_stage.state = StageStates.ANALYZING

        raw_results = current_stage.context.results
        execute_stages = current_stage.context.stages.get(StageTypes.EXECUTE)
        paths = current_stage.context.paths

        valid_states = [
            StageStates.EXECUTED, 
            StageStates.CHECKPOINTED, 
            StageStates.TEARDOWN_COMPLETE
        ]

        results_to_calculate = {}
        for stage_name in raw_results.keys():
            stage = execute_stages.get(stage_name)

            in_path = current_stage.name in paths.get(stage.name)

            if stage.state in valid_states and in_path:
                stage.state = StageStates.ANALYZING
                results_to_calculate[stage_name] = raw_results.get(stage_name)
        
        current_stage.raw_results = results_to_calculate

        if current_stage.timeout:
            summary = await asyncio.wait_for(current_stage.run(), timeout=current_stage.timeout)

        else:
            summary = await current_stage.run()

        current_stage.context.summaries.update(summary)

        current_stage.state = StageStates.ANALYZED

        next_stage.context = current_stage.context

        await logger.spinner.system.debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')
        await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')

    else:
        await logger.spinner.system.debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
        await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
        

async def analyze_to_checkpoint_transition(current_stage: Stage, next_stage: Stage):

    try:

        await analyze_transition(current_stage, next_stage)

    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR
    
    except Exception as stage_runtime_error:
        return StageExecutionError(current_stage, next_stage, str(stage_runtime_error)), StageTypes.ERROR

    next_stage.data = dict(current_stage.context.summaries)
    next_stage.previous_stage = current_stage.name

    current_stage = None
    
    return None, StageTypes.CHECKPOINT


async def analyze_to_submit_transition(current_stage: Stage, next_stage: Stage):

    try:

        await analyze_transition(current_stage, next_stage)

    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR
    
    except Exception as stage_runtime_error:
        print(traceback.format_exc())
        return StageExecutionError(current_stage, next_stage, str(stage_runtime_error)), StageTypes.ERROR

    current_stage = None

    return None, StageTypes.SUBMIT


async def analyze_to_wait_transition(current_stage: Stage, next_stage: Stage):

    try:

        await analyze_transition(current_stage, next_stage)

    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR
    
    except Exception as stage_runtime_error:
        return StageExecutionError(current_stage, next_stage, str(stage_runtime_error)), StageTypes.ERROR

    current_stage = None

    return None, StageTypes.WAIT



from typing import Dict
from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


class DatadogConfig(BaseModel):
    api_key: str
    app_key: str
    event_alert_type: str='info'
    device_name: str='hedra'
    priority: str='normal'
    custom_fields: Dict[str, str]={}
    reporter_type: ReporterTypes=ReporterTypes.Datadog


import asyncio
import traceback
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.types.stage_states import StageStates
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.transitions.exceptions import (
    StageExecutionError,
    StageTimeoutError
)
from hedra.logging import HedraLogger


async def submit_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    if current_stage.state == StageStates.INITIALIZED:

        await logger.spinner.system.debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')
        await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')

        current_stage.state = StageStates.SUBMITTING
        
        current_stage.summaries = current_stage.context.summaries

        if current_stage.timeout:
            await asyncio.wait_for(current_stage.run(), timeout=current_stage.timeout)
        
        else:
            await current_stage.run()

        current_stage.state = StageStates.SUBMITTED
        next_stage.state = StageStates.SUBMITTED

        await logger.spinner.system.debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')
        await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')

    else:
        await logger.spinner.system.debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
        await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
    
    next_stage.context = current_stage.context


async def submit_to_setup_transition(current_stage: Stage, next_stage: Stage):

    try:

        await submit_transition(current_stage, next_stage)
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR
    
    return None, StageTypes.SETUP


async def submit_to_optimize_transition(current_stage: Stage, next_stage: Stage):

    try:

        await submit_transition(current_stage, next_stage)
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    return None, StageTypes.OPTIMIZE


async def submit_to_execute_transition(current_stage: Stage, next_stage: Stage):

    try:

        await submit_transition(current_stage, next_stage)
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR
        
    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    return None, StageTypes.EXECUTE


async def submit_to_checkpoint_transition(current_stage: Stage, next_stage: Stage):

    try:

        await submit_transition(current_stage, next_stage)
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR
        
    return None, StageTypes.CHECKPOINT


async def submit_to_complete_transition(current_stage: Stage, next_stage: Stage):

    try:

        await submit_transition(current_stage, next_stage)
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    return None, StageTypes.COMPLETE


async def submit_to_submit_transition(current_stage: Stage, next_stage: Stage):

    try:

        await submit_transition(current_stage, next_stage)
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    return None, StageTypes.SUBMIT


async def submit_to_wait_transition(current_stage: Stage, next_stage: Stage):

    try:

        await submit_transition(current_stage, next_stage)
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    return None, StageTypes.WAIT



import asyncio
import dill
import threading
import os
from collections import defaultdict
from typing import Any, Dict, List, Union
from hedra.core.engines.client.config import Config
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.core.graphs.stages.optimize.optimization import Optimizer
from hedra.logging import HedraLogger
from hedra.core.personas import get_persona


def optimize_stage(serialized_config: str):

    logger = HedraLogger()
    logger.initialize()
    logger.filesystem.sync.create_logfile('hedra.optimize.log')

    thread_id = threading.current_thread().ident
    process_id = os.getpid()

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    optimization_config: Dict[str, Union[str, int, Any]] = dill.loads(serialized_config)

    graph_name: str = optimization_config.get('graph_name')
    graph_id: str = optimization_config.get('graph_id')
    source_stage_name: str = optimization_config.get('source_stage_name')
    source_stage_id: str = optimization_config.get('source_stage_id')
    execute_stage_name: str = optimization_config.get('execute_stage_name')
    execute_stage_config: Config = optimization_config.get('execute_stage_config')
    optimize_params: List[str] = optimization_config.get('optimize_params')
    optimize_iterations: int = optimization_config.get('optimizer_iterations')
    optimizer_algorithm: str = optimization_config.get('optimizer_algorithm')
    time_limit: int = optimization_config.get('time_limit')
    batch_size: int = optimization_config.get('execute_stage_batch_size')

    metadata_string = f'Graph - {graph_name}:{graph_id} - thread:{thread_id} - process:{process_id} - Stage: {source_stage_name}:{source_stage_id} - '

    execute_stage_hooks = defaultdict(list)
    for hook_name in optimization_config.get('execute_stage_hooks'):
        hook = registrar.all.get(hook_name)
        execute_stage_hooks[hook.hook_type].append(hook)


    logger.filesystem.sync['hedra.optimize'].info(f'{metadata_string} - Setting up Optimization')

    execute_stage_config.batch_size = batch_size

    persona = get_persona(execute_stage_config)
    persona.setup(execute_stage_hooks, metadata_string)

    optimizer = Optimizer({
        'graph_name': graph_name,
        'graph_id': graph_id,
        'source_stage_name': source_stage_name,
        'source_stage_id': source_stage_id,
        'params': optimize_params,
        'stage_name': execute_stage_name,
        'iterations': optimize_iterations,
        'algorithm': optimizer_algorithm,
        'persona': persona,
        'time_limit': time_limit
    })

    results = optimizer.optimize(loop)
    optimized_batch_size = results.get('optimized_batch_size')

    execute_stage_config.batch_size = optimized_batch_size
    loop.stop()
    loop.close()

    logger.filesystem.sync['hedra.optimize'].info(f'{optimizer.metadata_string} - Optimization complete')

    return {
        'stage': execute_stage_name,
        'config': execute_stage_config,
        'params': results
    }